import logging
import requests
from urllib.parse import urlencode

from astropy.time import Time, TimezoneInfo
from crispy_forms.layout import Column, Fieldset, Layout, Row
from django import forms
from django.core.cache import cache

from tom_alerts.alerts import GenericAlert, GenericBroker, GenericQueryForm
from tom_targets.models import Target

logger = logging.getLogger(__name__)

ALERCE_URL = 'https://alerce.online'
ALERCE_SEARCH_URL = 'https://api.alerce.online/ztf/v1'
ALERCE_CLASSES_URL = f'{ALERCE_SEARCH_URL}/classifiers'

SORT_CHOICES = [('None', 'None'),
                ('oid', 'Object ID'),
                ('probability', 'Classifier Probability'),
                ('ndet', 'Number of Detections'),
                ('firstmjd', 'First Detection'),
                ('lastmjd', 'Last Detection'),
                ('meanra', 'Mean Right Ascension'),
                ('meandec', 'Mean Declination')]

SORT_ORDER = [('None', 'None'),
              ('ASC', 'Ascending'),
              ('DESC', 'Descending')]


class ALeRCEQueryForm(GenericQueryForm):

    oid = forms.CharField(
        required=False,
        label='Object ID',
    )
    ndet = forms.IntegerField(
        required=False,
        label='Detections Lower',
        widget=forms.TextInput(attrs={'placeholder': 'Min number of detections'})
    )
    ranking = forms.IntegerField(
        required=False,
        label='Ranking',
        widget=forms.TextInput(attrs={'placeholder': 'Class ordering by probability'})
    )
    lc_classifier = forms.ChoiceField(
        required=False,
        label='Light Curve Class',
        choices=[],  # Choices are populated dynamically in the constructor
    )
    p_lc_classifier = forms.FloatField(
        required=False,
        label='Light Curve Classifier Probability'
    )
    stamp_classifier = forms.ChoiceField(
        required=False,
        label='Stamp Class',
        choices=[],  # Choices are populated dynamically in the constructor
    )
    p_stamp_classifier = forms.FloatField(
        required=False,
        label='Stamp Classifier Probability'
    )
    ra = forms.FloatField(
        required=False,
        label='RA',
        widget=forms.TextInput(attrs={'placeholder': 'RA (Degrees)'})
    )
    dec = forms.FloatField(
        required=False,
        label='Dec',
        widget=forms.TextInput(attrs={'placeholder': 'Dec (Degrees)'})
    )
    radius = forms.IntegerField(
        required=False,
        label='Search Radius',
        widget=forms.TextInput(attrs={'placeholder': 'Radius (Arcseconds)'})
    )
    firstmjd__gt = forms.FloatField(
        required=False,
        label='Min date of first detection ',
        widget=forms.TextInput(attrs={'placeholder': 'Date (MJD)'}),
        min_value=0.0
    )
    firstmjd__lt = forms.FloatField(
        required=False,
        label='Max date of first detection ',
        widget=forms.TextInput(attrs={'placeholder': 'Date (MJD)'}),
        min_value=0.0
    )
    lastmjd__gt = forms.FloatField(
        required=False,
        label='Min date of last detection',
        widget=forms.TextInput(attrs={'placeholder': 'Date (MJD)'}),
        min_value=0.0
    )
    lastmjd__lt = forms.FloatField(
        required=False,
        label='Max date of last detection',
        widget=forms.TextInput(attrs={'placeholder': 'Date (MJD)'}),
        min_value=0.0
    )
    order_by = forms.ChoiceField(
        choices=SORT_CHOICES,
        required=False,
        label='Sort By'
    )
    order_mode = forms.ChoiceField(
        choices=SORT_ORDER,
        required=False,
        label='Sort Order'
    )
    max_pages = forms.TypedChoiceField(
        choices=[(1, 1), (5, 5), (10, 10), (20, 20)],
        required=False,
        coerce=int,
        label='Maximum pages to retrieve'
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['lc_classifier'].choices = self._get_light_curve_classifier_choices()
        self.fields['stamp_classifier'].choices = self._get_stamp_classifier_choices()

        self.helper.layout = Layout(
            self.common_layout,
            'oid',
            Fieldset(
                'Classification Filters',
                Row(
                    Column('lc_classifier'),
                    Column('p_lc_classifier')
                ),
                Row(
                    Column('stamp_classifier'),
                    Column('p_stamp_classifier')
                )
            ),
            Fieldset(
                'Location Filters',
                Row(
                    Column('ra'),
                    Column('dec'),
                    Column('radius'),
                )
            ),
            Fieldset(
                'Time Filters',
                Row(
                    Column('firstmjd__gt'),
                    Column('firstmjd__lt'),
                ),
                Row(
                    Column('lastmjd__gt'),
                    Column('lastmjd__lt'),
                )
            ),
            Fieldset(
                'Other Filters',
                Row(
                    Column('ranking'),
                    Column('ndet')
                )
            ),
            Fieldset(
                'General Parameters',
                Row(
                    Column('order_by'),
                    Column('order_mode'),
                ),
                Row(
                    Column('max_pages')
                )
            ),
        )

    @staticmethod
    def _get_classifiers():
        cached_classifiers = cache.get('alerce_classifiers')

        if not cached_classifiers:
            response = requests.get(ALERCE_CLASSES_URL)
            response.raise_for_status()
            cached_classifiers = response.json()

        return cached_classifiers

    @staticmethod
    def _get_light_curve_classifier_choices():
        light_curve_classifiers = []
        for classifier in ALeRCEQueryForm._get_classifiers():
            if (any(x in classifier['classifier_name'] for x in ['transient', 'stochastic', 'periodic'])):
                classifier_name = classifier['classifier_name'].split('_')[-1]
                light_curve_classifiers += [(c, f'{c} - {classifier_name}') for c in classifier['classes']]

        return [(None, '')] + light_curve_classifiers

    @staticmethod
    def _get_stamp_classifier_choices():
        version = '0.0.0'
        stamp_classifiers = []

        for classifier in ALeRCEQueryForm._get_classifiers():
            if classifier['classifier_name'] == 'stamp_classifier':
                new_version = classifier['classifier_version'].split('_')[-1]
                if new_version > version:
                    stamp_classifiers = [(c, c) for c in classifier['classes']]

        return [(None, '')] + stamp_classifiers

    def clean_max_pages(self):
        max_pages = self.cleaned_data['max_pages']
        if not max_pages:
            max_pages = 1
        return max_pages

    def clean(self):
        cleaned_data = super().clean()

        # Ensure that all cone search fields are present
        if (any(cleaned_data[k] for k in ['ra', 'dec', 'radius'])
                and not all(cleaned_data[k] for k in ['ra', 'dec', 'radius'])):
            raise forms.ValidationError('All of RA, Dec, and Search Radius must be included to execute a cone search.')

        # Ensure that only one classification set is filled in
        if (any(cleaned_data.get(k) for k in ['lc_classifier', 'p_lc_classifier'])
                and any(cleaned_data.get(k) for k in ['stamp_classifier', 'p_stamp_classifier'])):
            raise forms.ValidationError('Only one of either light curve or stamp classification may be used as a '
                                        'filter.')

        return cleaned_data


class ALeRCEBroker(GenericBroker):
    name = 'ALeRCE'
    form = ALeRCEQueryForm

    def _clean_classifier_parameters(self, parameters):
        classifier_parameters = []
        class_type = ''
        if parameters['stamp_classifier']:
            class_type = 'stamp_classifier'
        elif parameters['lc_classifier']:
            class_type = 'lc_classifier'

        if class_type:
            classifier_parameters.append(('classifier', class_type))
        if class_type in parameters and parameters[class_type] is not None:
            classifier_parameters.append(('class', parameters[class_type]))
        if f'p_{class_type}' in parameters and parameters[f'p_{class_type}'] is not None:
            classifier_parameters.append(('probability', parameters[f'p_{class_type}']))

        return classifier_parameters

    def _clean_coordinate_parameters(self, parameters):
        if all([parameters['ra'], parameters['dec'], parameters['radius']]):
            return [
                ('ra', parameters['ra']),
                ('dec', parameters['dec']),
                ('radius', parameters['radius'])
            ]
        else:
            return []

    def _clean_date_parameters(self, parameters):
        dates = []

        dates += [('firstmjd', v) for k, v in parameters.items() if 'firstmjd' in k and v]
        dates += [('lastmjd', v) for k, v in parameters.items() if 'lastmjd' in k and v]

        return dates

    def _clean_parameters(self, parameters):
        payload = [
            (k, v) for k, v in parameters.items() if k in ['oid', 'ndet', 'ranking', 'order_by', 'order_mode'] and v
        ]

        payload += [
            ('page', parameters.get('page', 1)),
            ('page_size', 20),
        ]

        payload += self._clean_classifier_parameters(parameters)

        payload += self._clean_coordinate_parameters(parameters)

        payload += self._clean_date_parameters(parameters)

        return payload

    def _request_alerts(self, parameters):
        payload = self._clean_parameters(parameters)
        logger.log(msg=f'Fetching alerts from ALeRCE with payload {payload}', level=logging.INFO)
        args = urlencode(self._clean_parameters(parameters))
        response = requests.get(f'{ALERCE_SEARCH_URL}/objects/?count=false&{args}')
        response.raise_for_status()
        return response.json()

    def fetch_alerts(self, parameters):
        response = self._request_alerts(parameters)
        alerts = response['items']
        if len(alerts) > 0 and response['page'] < parameters.get('max_pages', 1):
            parameters['page'] = response.get('page') + 1
            alerts += self.fetch_alerts(parameters)
        return iter(alerts)

    def fetch_alert(self, id):
        """
        The response for a single alert is as follows:

        {
            'oid':'ZTF20acnsdjd',
            ...
            'firstmjd':59149.1119328998,
            ...
        }
        """
        response = requests.get(f'{ALERCE_SEARCH_URL}/objects/{id}')
        response.raise_for_status()
        return response.json()

    def to_target(self, alert):
        return Target.objects.create(
            name=alert['oid'],
            type='SIDEREAL',
            ra=alert['meanra'],
            dec=alert['meandec']
        )

    # TODO: the generic alert is clearly not sufficient for ALeRCE's classifications
    def to_generic_alert(self, alert):
        if alert['lastmjd']:
            timestamp = Time(alert['lastmjd'], format='mjd', scale='utc').to_datetime(timezone=TimezoneInfo())
        else:
            timestamp = ''
        url = f'{ALERCE_URL}/object/{alert["oid"]}'

        mag = None  # mag is no longer returned in the object list

        score = alert['probability']

        return GenericAlert(
            timestamp=timestamp,
            url=url,
            id=alert['oid'],
            name=alert['oid'],
            ra=alert['meanra'],
            dec=alert['meandec'],
            mag=mag,
            score=score
        )



from datetime import datetime, timedelta
from dateutil.parser import parse
import logging

from tom_observations.cadence import BaseCadenceForm, CadenceStrategy
from tom_observations.models import ObservationRecord
from tom_observations.facility import get_service_class

logger = logging.getLogger(__name__)


class ResumeCadenceAfterFailureForm(BaseCadenceForm):
    pass


class ResumeCadenceAfterFailureStrategy(CadenceStrategy):
    """The ResumeCadenceAfterFailureStrategy chooses when to submit the next observation based on the success of the
    previous observation. If the observation is successful, it submits a new one on the same cadence--that is, if the
    cadence is every three days, it will submit the next observation three days in the future. If the observations
    fails, it will submit the next observation immediately, and follow the same decision tree based on the success
    of the subsequent observation.

    In order to properly subclass this CadenceStrategy, one should override ``update_observation_payload``.

    This strategy requires the DynamicCadence to have a parameter ``cadence_frequency``."""

    name = 'Resume Cadence After Failure'
    description = """This strategy schedules one observation in the cadence at a time. If the observation fails, it
                     re-submits the observation until it succeeds. If it succeeds, it submits the next observation on
                     the same cadence."""
    form = ResumeCadenceAfterFailureForm

    def update_observation_payload(self, observation_payload):
        """
        :param observation_payload: form parameters for facility observation form
        :type observation_payload: dict
        """
        return observation_payload

    def run(self):
        # gets the most recent observation because the next observation is just going to modify these parameters
        last_obs = self.dynamic_cadence.observation_group.observation_records.order_by('-created').first()

        # Make a call to the facility to get the current status of the observation
        facility = get_service_class(last_obs.facility)()
        facility.update_observation_status(last_obs.observation_id)  # Updates the DB record
        last_obs.refresh_from_db()  # Gets the record updates

        # Boilerplate to get necessary properties for future calls
        start_keyword, end_keyword = facility.get_start_end_keywords()
        observation_payload = last_obs.parameters

        # Cadence logic
        # If the observation hasn't finished, do nothing
        if not last_obs.terminal:
            return
        elif last_obs.failed:  # If the observation failed
            # Submit next observation to be taken as soon as possible with the same window length
            window_length = parse(observation_payload[end_keyword]) - parse(observation_payload[start_keyword])
            observation_payload[start_keyword] = datetime.now().isoformat()
            observation_payload[end_keyword] = (parse(observation_payload[start_keyword]) + window_length).isoformat()
        else:  # If the observation succeeded
            # Advance window normally according to cadence parameters
            observation_payload = self.advance_window(
                observation_payload, start_keyword=start_keyword, end_keyword=end_keyword
            )

        observation_payload = self.update_observation_payload(observation_payload)

        # Submission of the new observation to the facility
        obs_type = last_obs.parameters.get('observation_type')
        form = facility.get_form(obs_type)(observation_payload)
        if form.is_valid():
            observation_ids = facility.submit_observation(form.observation_payload())
        else:
            logger.error(msg=f'Unable to submit next cadenced observation: {form.errors}')
            raise Exception(f'Unable to submit next cadenced observation: {form.errors}')

        # Creation of corresponding ObservationRecord objects for the observations
        new_observations = []
        for observation_id in observation_ids:
            # Create Observation record
            record = ObservationRecord.objects.create(
                target=last_obs.target,
                facility=facility.name,
                parameters=observation_payload,
                observation_id=observation_id
            )
            # Add ObservationRecords to the DynamicCadence
            self.dynamic_cadence.observation_group.observation_records.add(record)
            self.dynamic_cadence.observation_group.save()
            new_observations.append(record)

        # Update the status of the ObservationRecords in the DB
        for obsr in new_observations:
            facility = get_service_class(obsr.facility)()
            facility.update_observation_status(obsr.observation_id)

        return new_observations

    def advance_window(self, observation_payload, start_keyword='start', end_keyword='end'):
        cadence_frequency = self.dynamic_cadence.cadence_parameters.get('cadence_frequency')
        if not cadence_frequency:
            raise Exception(f'The {self.name} strategy requires a cadence_frequency cadence_parameter.')
        advance_window_hours = cadence_frequency
        window_length = parse(observation_payload[end_keyword]) - parse(observation_payload[start_keyword])

        new_start = parse(observation_payload[start_keyword]) + timedelta(hours=advance_window_hours)
        if new_start < datetime.now():  # Ensure that the new window isn't in the past
            new_start = datetime.now()
        new_end = new_start + window_length
        observation_payload[start_keyword] = new_start.isoformat()
        observation_payload[end_keyword] = new_end.isoformat()

        return observation_payload



from django.db.models import Count

import csv
from .models import Target, TargetExtra, TargetName
from io import StringIO


# NOTE: This saves locally. To avoid this, create file buffer.
# referenced https://www.codingforentrepreneurs.com/blog/django-queryset-to-csv-files-datasets/
def export_targets(qs):
    """
    Exports all the specified targets into a csv file in folder csvTargetFiles
    NOTE: This saves locally. To avoid this, create file buffer.

    :param qs: List of targets to export
    :type qs: QuerySet

    :returns: String buffer of exported targets
    :rtype: StringIO
    """
    qs_pk = [data['id'] for data in qs]
    data_list = list(qs)
    target_fields = [field.name for field in Target._meta.get_fields()]
    target_extra_fields = list({field.key for field in TargetExtra.objects.filter(target__in=qs_pk)})
    # Gets the count of the target names for the target with the most aliases in the database
    # This is to construct enough row headers of format "name2, name3, name4, etc" for exporting aliases
    # The alias headers are then added to the set of fields for export
    aliases = TargetName.objects.filter(target__in=qs_pk).values('target_id').annotate(count=Count('target_id'))
    max_alias_count = 0
    if aliases:
        max_alias_count = max([alias['count'] for alias in aliases])
    all_fields = target_fields + target_extra_fields + [f'name{index+1}' for index in range(1, max_alias_count+1)]
    for key in ['id', 'targetlist', 'dataproduct', 'observationrecord', 'reduceddatum', 'aliases', 'targetextra']:
        all_fields.remove(key)

    file_buffer = StringIO()
    writer = csv.DictWriter(file_buffer, fieldnames=all_fields)
    writer.writeheader()
    for target_data in data_list:
        extras = list(TargetExtra.objects.filter(target_id=target_data['id']))
        names = list(TargetName.objects.filter(target_id=target_data['id']))
        for e in extras:
            target_data[e.key] = e.value
        name_index = 2
        for name in names:
            target_data[f'name{str(name_index)}'] = name.name
            name_index += 1
        del target_data['id']  # do not export 'id'
        writer.writerow(target_data)
    return file_buffer


def import_targets(targets):
    """
    Imports a set of targets into the TOM and saves them to the database.

    :param targets: String buffer of targets
    :type targets: StringIO

    :returns: dictionary of successfully imported targets, as well errors
    :rtype: dict
    """
    # TODO: Replace this with an in memory iterator
    targetreader = csv.DictReader(targets, dialect=csv.excel)
    targets = []
    errors = []
    base_target_fields = [field.name for field in Target._meta.get_fields()]
    for index, row in enumerate(targetreader):
        # filter out empty values in base fields, otherwise converting empty string to float will throw error
        row = {k: v for (k, v) in row.items() if not (k in base_target_fields and not v)}
        target_extra_fields = []
        target_names = []
        target_fields = {}
        for k in row:
            # All fields starting with 'name' (e.g. name2, name3) that aren't literally 'name' will be added as
            # TargetNames
            if k != 'name' and k.startswith('name'):
                target_names.append(row[k])
            elif k not in base_target_fields:
                target_extra_fields.append((k, row[k]))
            else:
                target_fields[k] = row[k]
        for extra in target_extra_fields:
            row.pop(extra[0])
        try:
            target = Target.objects.create(**target_fields)
            for extra in target_extra_fields:
                TargetExtra.objects.create(target=target, key=extra[0], value=extra[1])
            for name in target_names:
                if name:
                    TargetName.objects.create(target=target, name=name)
            targets.append(target)
        except Exception as e:
            error = 'Error on line {0}: {1}'.format(index + 2, str(e))
            errors.append(error)

    return {'targets': targets, 'errors': errors}



import mimetypes
import numpy as np

from datetime import datetime

from astropy import units
from astropy.io import fits, ascii
from astropy.time import Time
from astropy.wcs import WCS
from specutils import Spectrum1D

from tom_dataproducts.data_processor import DataProcessor
from tom_dataproducts.exceptions import InvalidFileFormatException
from tom_dataproducts.processors.data_serializers import SpectrumSerializer
from tom_observations.facility import get_service_class, get_service_classes


class SpectroscopyProcessor(DataProcessor):

    DEFAULT_WAVELENGTH_UNITS = units.angstrom
    DEFAULT_FLUX_CONSTANT = units.erg / units.cm ** 2 / units.second / units.angstrom

    def process_data(self, data_product):
        """
        Routes a spectroscopy processing call to a method specific to a file-format, then serializes the returned data.

        :param data_product: Spectroscopic DataProduct which will be processed into the specified format for database
        ingestion
        :type data_product: DataProduct

        :returns: python list of 2-tuples, each with a timestamp and corresponding data
        :rtype: list
        """

        mimetype = mimetypes.guess_type(data_product.data.path)[0]
        if mimetype in self.FITS_MIMETYPES:
            spectrum, obs_date = self._process_spectrum_from_fits(data_product)
        elif mimetype in self.PLAINTEXT_MIMETYPES:
            spectrum, obs_date = self._process_spectrum_from_plaintext(data_product)
        else:
            raise InvalidFileFormatException('Unsupported file type')

        serialized_spectrum = SpectrumSerializer().serialize(spectrum)

        return [(obs_date, serialized_spectrum)]

    def _process_spectrum_from_fits(self, data_product):
        """
        Processes the data from a spectrum from a fits file into a Spectrum1D object, which can then be serialized and
        stored as a ReducedDatum for further processing or display. File is read using specutils as specified in the
        below documentation.
        # https://specutils.readthedocs.io/en/doc-testing/specutils/read_fits.html

        :param data_product: Spectroscopic DataProduct which will be processed into a Spectrum1D
        :type data_product: tom_dataproducts.models.DataProduct

        :returns: Spectrum1D object containing the data from the DataProduct
        :rtype: specutils.Spectrum1D

        :returns: Datetime of observation, if it is in the header and the file is from a supported facility, current
            datetime otherwise
        :rtype: AstroPy.Time
        """

        flux, header = fits.getdata(data_product.data.path, header=True)

        for facility_class in get_service_classes():
            facility = get_service_class(facility_class)()
            if facility.is_fits_facility(header):
                flux_constant = facility.get_flux_constant()
                date_obs = facility.get_date_obs_from_fits_header(header)
                break
        else:
            flux_constant = self.DEFAULT_FLUX_CONSTANT
            date_obs = datetime.now()

        dim = len(flux.shape)
        if dim == 3:
            flux = flux[0, 0, :]
        elif flux.shape[0] == 2:
            flux = flux[0, :]
        flux = flux * flux_constant

        header['CUNIT1'] = 'Angstrom'
        wcs = WCS(header=header, naxis=1)

        spectrum = Spectrum1D(flux=flux, wcs=wcs)

        return spectrum, Time(date_obs).to_datetime()

    def _process_spectrum_from_plaintext(self, data_product):
        """
        Processes the data from a spectrum from a plaintext file into a Spectrum1D object, which can then be serialized
        and stored as a ReducedDatum for further processing or display. File is read using astropy as specified in
        the below documentation. The file is expected to be a multi-column delimited file, with headers for wavelength
        and flux. The file also requires comments containing, at minimum, 'DATE-OBS: [value]', where value is an
        Astropy Time module-readable date. It can optionally contain 'FACILITY: [value]', where the facility is a string
        matching the name of a valid facility in the TOM.
        # http://docs.astropy.org/en/stable/io/ascii/read.html

        Parameters
        ----------
        :param data_product: Spectroscopic DataProduct which will be processed into a Spectrum1D
        :type data_product: tom_dataproducts.models.DataProduct

        :returns: Spectrum1D object containing the data from the DataProduct
        :rtype: specutils.Spectrum1D

        :returns: Datetime of observation, if it is in the comments and the file is from a supported facility, current
            datetime otherwise
        :rtype: AstroPy.Time
        """

        data = ascii.read(data_product.data.path)
        if len(data) < 1:
            raise InvalidFileFormatException('Empty table or invalid file type')
        facility_name = None
        date_obs = datetime.now()
        comments = data.meta.get('comments', [])

        for comment in comments:
            if 'date-obs' in comment.lower():
                date_obs = comment.split(':')[1].strip()
            if 'facility' in comment.lower():
                facility_name = comment.split(':')[1].strip()

        facility = get_service_class(facility_name)() if facility_name else None
        wavelength_units = facility.get_wavelength_units() if facility else self.DEFAULT_WAVELENGTH_UNITS
        flux_constant = facility.get_flux_constant() if facility else self.DEFAULT_FLUX_CONSTANT

        spectral_axis = np.array(data['wavelength']) * wavelength_units
        flux = np.array(data['flux']) * flux_constant
        spectrum = Spectrum1D(flux=flux, spectral_axis=spectral_axis)

        return spectrum, Time(date_obs).to_datetime()



import pytz
from datetime import datetime

from django.contrib.auth.models import User, Group
from django.contrib.messages import get_messages
from django.contrib.messages.constants import SUCCESS, WARNING
from django.test import TestCase, override_settings
from django.urls import reverse

from .factories import SiderealTargetFactory, NonSiderealTargetFactory, TargetGroupingFactory, TargetNameFactory
from tom_targets.models import Target, TargetExtra, TargetList, TargetName
from tom_targets.utils import import_targets
from guardian.shortcuts import assign_perm


class TestTargetListUserPermissions(TestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser')
        self.user2 = User.objects.create(username='unauthorized')
        self.st1 = SiderealTargetFactory.create()
        self.st2 = SiderealTargetFactory.create()

        assign_perm('tom_targets.view_target', self.user, self.st1)
        assign_perm('tom_targets.view_target', self.user, self.st2)
        assign_perm('tom_targets.view_target', self.user2, self.st2)

    def test_list_targets(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('targets:list'))
        self.assertContains(response, self.st1.name)
        self.assertContains(response, self.st2.name)

    def test_list_targets_limited_permissions(self):
        self.client.force_login(self.user2)
        response = self.client.get(reverse('targets:list'))
        self.assertContains(response, self.st2.name)
        self.assertNotContains(response, self.st1.name)


class TestTargetListGroupPermissions(TestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser')
        self.user2 = User.objects.create(username='unauthorized')
        self.st1 = SiderealTargetFactory.create()
        self.st2 = SiderealTargetFactory.create()
        self.group1 = Group.objects.create(name='group1')
        self.group2 = Group.objects.create(name='group2')
        self.group1.user_set.add(self.user)
        self.group2.user_set.add(self.user)
        self.group2.user_set.add(self.user2)

        assign_perm('tom_targets.view_target', self.group1, self.st1)
        assign_perm('tom_targets.view_target', self.group2, self.st2)

    def test_list_targets(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('targets:list'))
        self.assertContains(response, self.st1.name)
        self.assertContains(response, self.st2.name)

    def test_list_targets_limited_permissions(self):
        self.client.force_login(self.user2)
        response = self.client.get(reverse('targets:list'))
        self.assertContains(response, self.st2.name)
        self.assertNotContains(response, self.st1.name)


# Because the target detail page has a templatetag that tries to get the facility status, these tests fail without
# network. While the preferred solution would be to create a mock facility class, in order to avoid any potential
# circular imports, we're simply disabling the facility classes for these tests. This can be revisited if need be at a
# future time, but currently the target tests don't do anything with ObservationRecords anyway.
@override_settings(TOM_FACILITY_CLASSES=[])
class TestTargetDetail(TestCase):
    def setUp(self):
        user = User.objects.create(username='testuser')
        self.client.force_login(user)
        self.st = SiderealTargetFactory.create(ra=123.456, dec=-32.1)
        self.nst = NonSiderealTargetFactory.create()
        assign_perm('tom_targets.view_target', user, self.st)
        assign_perm('tom_targets.view_target', user, self.nst)

    def test_sidereal_target_detail(self):
        response = self.client.get(reverse('targets:detail', kwargs={'pk': self.st.id}))
        self.assertContains(response, self.st.id)

    def test_non_sidereal_target_detail(self):
        response = self.client.get(reverse('targets:detail', kwargs={'pk': self.nst.id}))
        self.assertContains(response, self.nst.id)

    @override_settings(EXTRA_FIELDS=[
        {'name': 'somefield', 'type': 'string'},
        {'name': 'hiddenfield', 'type': 'string', 'hidden': True},
    ])
    def test_extra_fields(self):
        self.st.save(extras={'somefield': 'somevalue', 'hiddenfield': 'hiddenvalue'})
        response = self.client.get(reverse('targets:detail', kwargs={'pk': self.st.id}))
        self.assertContains(response, 'somevalue')
        self.assertNotContains(response, 'hiddenvalue')

    def test_target_bad_permissions(self):
        other_user = User.objects.create(username='otheruser')
        self.client.force_login(other_user)
        response = self.client.get(reverse('targets:detail', kwargs={'pk': self.st.id}), follow=True)
        self.assertRedirects(response, '{}?next=/targets/{}/'.format(reverse('login'), self.st.id))
        self.assertContains(response, 'You do not have permission to access this page')


class TestTargetNameSearch(TestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser')
        self.user2 = User.objects.create(username='testuser2')

        self.st1 = SiderealTargetFactory.create(name='testtarget1')
        self.st2 = SiderealTargetFactory.create(name='testtarget2')
        self.st3 = SiderealTargetFactory.create(name='testtarget3')

        assign_perm('tom_targets.view_target', self.user, self.st1)
        assign_perm('tom_targets.view_target', self.user2, self.st1)
        assign_perm('tom_targets.view_target', self.user, self.st2)
        assign_perm('tom_targets.view_target', self.user, self.st3)

    def test_search_one_result(self):
        """Test that a search with one result returns the target detail page."""
        self.client.force_login(self.user)
        response = self.client.get(reverse('targets:name-search', kwargs={'name': self.st1.name}), follow=True)
        self.assertRedirects(response, reverse('targets:detail', kwargs={'pk': self.st1.id}))
        self.assertContains(response, self.st1.name)

    def test_search_no_results(self):
        """Test that a search with no results returns the target list page."""
        self.client.force_login(self.user)
        response = self.client.get(reverse('targets:name-search', kwargs={'name': 'fakename'}), follow=True)
        self.assertRedirects(response, reverse('targets:list') + '?name=fakename')
        self.assertNotContains(response, self.st1.name)
        self.assertNotContains(response, self.st2.name)

    def test_search_multiple_results(self):
        """Test that a search with multiple results returns the target list page."""
        self.client.force_login(self.user)
        response = self.client.get(reverse('targets:name-search', kwargs={'name': 'testtarget'}), follow=True)
        self.assertRedirects(response, reverse('targets:list') + '?name=testtarget')
        self.assertContains(response, self.st1.name)
        self.assertContains(response, self.st3.name)

    def test_search_one_result_unauthorized(self):
        """Test that a search with one result that the user is not allowed to view returns an empty target list page."""
        self.client.force_login(self.user2)
        response = self.client.get(reverse('targets:name-search', kwargs={'name': 'testtarget3'}), follow=True)
        self.assertRedirects(response, reverse('targets:list') + '?name=testtarget3')
        self.assertContains(response, 'No targets match those filters.')

    def test_search_multiple_results_unauthorized(self):
        """Test that a search with multiple results returns the target list page, but without the targets that
           the user is not allowed to view."""
        assign_perm('tom_targets.view_target', self.user2, self.st2)
        self.client.force_login(self.user2)
        response = self.client.get(reverse('targets:name-search', kwargs={'name': 'testtarget'}), follow=True)
        self.assertRedirects(response, reverse('targets:list') + '?name=testtarget')
        self.assertContains(response, self.st1.name)
        self.assertNotContains(response, self.st3.name)

    def test_search_one_result_authorized(self):
        """Test that a search with only one result that the user is allowed to view returns the target detail page."""
        self.client.force_login(self.user2)
        response = self.client.get(reverse('targets:name-search', kwargs={'name': 'testtarget'}), follow=True)
        self.assertRedirects(response, reverse('targets:detail', kwargs={'pk': self.st1.id}))
        self.assertContains(response, self.st1.name)


@override_settings(TOM_FACILITY_CLASSES=[])
class TestTargetCreate(TestCase):
    def setUp(self):
        user = User.objects.create(username='testuser')
        self.client.force_login(user)
        self.group = Group.objects.create(name='agroup')
        user.groups.add(self.group)
        user.save()

    def test_target_create_form(self):
        response = self.client.get(reverse('targets:create'))
        self.assertContains(response, Target.SIDEREAL)
        self.assertContains(response, Target.NON_SIDEREAL)

    def test_create_target(self):
        target_data = {
            'name': 'test_target_name',
            'type': Target.SIDEREAL,
            'ra': 123.456,
            'dec': -32.1,
            'groups': [self.group.id],
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': '',
            'targetextra_set-0-value': '',
            'aliases-TOTAL_FORMS': 1,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        response = self.client.post(reverse('targets:create'), data=target_data, follow=True)
        self.assertContains(response, target_data['name'])
        self.assertTrue(Target.objects.filter(name=target_data['name']).exists())

    def test_create_target_sexigesimal(self):
        """
        Using coordinates for Messier 1
        """
        target_data = {
            'name': 'test_target_name',
            'type': Target.SIDEREAL,
            'ra': '05:34:31.94',
            'dec': '+22:00:52.2',
            'groups': [self.group.id],
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': '',
            'targetextra_set-0-value': '',
            'aliases-TOTAL_FORMS': 1,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        response = self.client.post(reverse('targets:create'), data=target_data, follow=True)
        self.assertContains(response, target_data['name'])
        target = Target.objects.get(name=target_data['name'])
        # Coordinates according to simbad
        self.assertAlmostEqual(target.ra, 83.63308, places=4)
        self.assertAlmostEqual(target.dec, 22.0145, places=4)

    def test_create_target_with_tags(self):
        """
        Using coordinates for Messier 1
        """
        target_data = {
            'name': 'test_target',
            'type': Target.SIDEREAL,
            'ra': '05:34:31.94',
            'dec': '+22:00:52.2',
            'groups': [self.group.id],
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': 'category',
            'targetextra_set-0-value': 'type2',
            'aliases-TOTAL_FORMS': 1,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        response = self.client.post(reverse('targets:create'), data=target_data, follow=True)
        self.assertContains(response, target_data['name'])
        target = Target.objects.get(name=target_data['name'])
        self.assertTrue(target.targetextra_set.filter(key='category', value='type2').exists())

    @override_settings(EXTRA_FIELDS=[
        {'name': 'wins', 'type': 'number'},
        {'name': 'checked', 'type': 'boolean'},
        {'name': 'birthdate', 'type': 'datetime'},
        {'name': 'author', 'type': 'string'}

    ])
    def test_create_target_with_extra_fields(self):
        target_data = {
            'name': 'extra_field_target',
            'type': Target.SIDEREAL,
            'ra': 113.456,
            'dec': -22.1,
            'wins': 50.0,
            'checked': True,
            'birthdate': datetime(year=2019, month=2, day=14),
            'author': 'Dr. Suess',
            'groups': [self.group.id],
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': '',
            'targetextra_set-0-value': '',
            'aliases-TOTAL_FORMS': 1,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        response = self.client.post(reverse('targets:create'), data=target_data, follow=True)
        self.assertContains(response, target_data['name'])
        target = Target.objects.get(name=target_data['name'])
        self.assertTrue(TargetExtra.objects.filter(target=target, key='wins', value='50.0').exists())
        self.assertEqual(
            TargetExtra.objects.get(target=target, key='wins').typed_value('number'),
            50.0
        )
        self.assertEqual(
            TargetExtra.objects.get(target=target, key='checked').typed_value('boolean'),
            True
        )
        self.assertEqual(
            TargetExtra.objects.get(target=target, key='birthdate').typed_value('datetime'),
            datetime(year=2019, month=2, day=14, tzinfo=pytz.UTC)
        )
        self.assertEqual(
            TargetExtra.objects.get(target=target, key='author').typed_value('string'),
            'Dr. Suess'
        )
        # Check extra_fields property converts values to the correct type
        self.assertIsInstance(target.extra_fields, dict)
        self.assertIsInstance(target.extra_fields['wins'], float)
        self.assertIsInstance(target.extra_fields['checked'], bool)
        self.assertIsInstance(target.extra_fields['birthdate'], datetime)
        self.assertIsInstance(target.extra_fields['author'], str)

    def test_target_save_programmatic_extras(self):
        target = SiderealTargetFactory.create()
        target.save(extras={'foo': 5})
        self.assertTrue(TargetExtra.objects.filter(target=target, key='foo', value='5').exists())

    def test_non_sidereal_required_fields(self):
        base_data = {
            'name': 'nonsidereal_target',
            'identifier': 'nonsidereal_identifier',
            'type': Target.NON_SIDEREAL,
            'epoch_of_elements': 100,
            'lng_asc_node': 100,
            'arg_of_perihelion': 100,
            'eccentricity': 100,
            'mean_anomaly': 100,
            'inclination': 100,
            'semimajor_axis': 100,
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': '',
            'targetextra_set-0-value': '',
            'aliases-TOTAL_FORMS': 1,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        create_url = reverse('targets:create') + '?type=NON_SIDEREAL'

        # Make data for a major planet scheme: missing 'mean_daily_motion'
        maj_planet_data = dict(**base_data, scheme='JPL_MAJOR_PLANET')
        response = self.client.post(create_url, data=maj_planet_data, follow=True)
        errors = response.context['form'].errors
        self.assertEqual(set(errors.keys()), {'__all__'})
        messages = errors['__all__']
        self.assertEqual(len(messages), 1)
        self.assertTrue(messages[0].startswith("Scheme 'JPL Major Planet' requires fields"))
        self.assertIn('Daily Motion', messages[0])

        # Use the same data for minor planet: should be no errors
        min_planet_data = dict(**base_data, scheme='MPC_MINOR_PLANET')
        response = self.client.post(create_url, data=min_planet_data, follow=True)
        errors = response.context['form'].errors
        self.assertEqual(errors, {})

    def test_create_form_failure(self):
        """
        If a failure occurs when creating a non-sidereal target, make sure the
        user is shown the non-sidereal form (not the sidereal one)
        """
        target_data = {
            'name': 'test_target',
            'type': Target.NON_SIDEREAL,
            'scheme': 'MPC_MINOR_PLANET',
            # Miss out a load of required fields...
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': '',
            'targetextra_set-0-value': '',
            'aliases-TOTAL_FORMS': 1,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        response = self.client.post(reverse('targets:create'), data=target_data)
        self.assertEqual(response.context['form'].initial['type'], Target.NON_SIDEREAL)

    def test_create_targets_with_multiple_names(self):
        target_data = {
            'name': 'multiple_names_target',
            'type': Target.SIDEREAL,
            'ra': 113.456,
            'dec': -22.1,
            'groups': [self.group.id],
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': '',
            'targetextra_set-0-value': '',
            'aliases-TOTAL_FORMS': 2,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        names = ['John', 'Doe']
        for i, name in enumerate(names):
            target_data[f'aliases-{i}-name'] = name
        response = self.client.post(reverse('targets:create'), data=target_data, follow=True)
        self.assertContains(response, target_data['name'])

        target = Target.objects.get(name=target_data['name'])
        for target_name in names:
            self.assertTrue(TargetName.objects.filter(target=target, name=target_name).exists())

    def test_create_targets_with_conflicting_names(self):
        target_data = {
            'name': 'multiple_names_target',
            'type': Target.SIDEREAL,
            'ra': 113.456,
            'dec': -22.1,
            'groups': [self.group.id],
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': '',
            'targetextra_set-0-value': '',
            'aliases-TOTAL_FORMS': 2,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        names = ['John', 'Doe']
        for i, name in enumerate(names):
            target_data[f'aliases-{i}-name'] = name
        self.client.post(reverse('targets:create'), data=target_data, follow=True)
        second_response = self.client.post(reverse('targets:create'), data=target_data, follow=True)
        self.assertContains(second_response, 'Target with this Name already exists')

    def test_create_targets_with_conflicting_aliases(self):
        target_data = {
            'name': 'multiple_names_target',
            'type': Target.SIDEREAL,
            'ra': 113.456,
            'dec': -22.1,
            'groups': [self.group.id],
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': '',
            'targetextra_set-0-value': '',
            'aliases-TOTAL_FORMS': 2,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
        }
        names = ['John', 'Doe']
        for i, name in enumerate(names):
            target_data[f'aliases-{i}-name'] = name
        self.client.post(reverse('targets:create'), data=target_data, follow=True)
        target_data['name'] = 'multiple_names_target2'
        second_response = self.client.post(reverse('targets:create'), data=target_data, follow=True)
        self.assertContains(second_response, 'Target name with this Alias already exists.')


class TestTargetUpdate(TestCase):
    def setUp(self):
        self.form_data = {
            'name': 'testtarget',
            'type': Target.SIDEREAL,
            'ra': 113.456,
            'dec': -22.1
        }
        user = User.objects.create(username='testuser')
        self.target = Target.objects.create(**self.form_data)
        assign_perm('tom_targets.change_target', user, self.target)
        self.client.force_login(user)

    def test_valid_update(self):
        self.form_data.update({
            'targetextra_set-TOTAL_FORMS': 1,
            'targetextra_set-INITIAL_FORMS': 0,
            'targetextra_set-MIN_NUM_FORMS': 0,
            'targetextra_set-MAX_NUM_FORMS': 1000,
            'targetextra_set-0-key': 'redshift',
            'targetextra_set-0-value': '3',
            'aliases-TOTAL_FORMS': 1,
            'aliases-INITIAL_FORMS': 0,
            'aliases-MIN_NUM_FORMS': 0,
            'aliases-MAX_NUM_FORMS': 1000,
            'aliases-0-name': 'testtargetname2'
        })
        self.client.post(reverse('targets:update', kwargs={'pk': self.target.id}), data=self.form_data)
        self.target.refresh_from_db()
        self.assertTrue(self.target.targetextra_set.filter(key='redshift').exists())
        self.assertTrue(self.target.aliases.filter(name='testtargetname2').exists())


class TestTargetImport(TestCase):
    def setUp(self):
        user = User.objects.create(username='testuser')
        self.client.force_login(user)

    def test_import_csv(self):
        csv = [
            'name,type,ra,dec',
            'm13,SIDEREAL,250.421,36.459',
            'm27,SIDEREAL,299.901,22.721'
        ]
        result = import_targets(csv)
        self.assertEqual(len(result['targets']), 2)

    @override_settings(EXTRA_FIELDS=[{'name': 'redshift', 'type': 'number'}])
    def test_import_with_extra(self):
        csv = [
            'name,type,ra,dec,redshift',
            'm13,SIDEREAL,250.421,36.459,5',
            'm27,SIDEREAL,299.901,22.721,5'
        ]
        result = import_targets(csv)
        self.assertEqual(len(result['targets']), 2)
        for target in result['targets']:
            self.assertTrue(TargetExtra.objects.filter(target=target, key='redshift', value='5').exists())

    def test_import_csv_with_multiple_names(self):
        csv = [
            'name,type,ra,dec,name1,name2',
            'm13,SIDEREAL,250.421,36.459,Tom,Joe',
            'm27,SIDEREAL,299.901,22.721,John,Doe'
        ]
        result = import_targets(csv)
        self.assertEqual(len(result['targets']), 2)
        aliases = {'m13': 'Tom,Joe', 'm27': 'John,Doe'}
        for target_name in aliases:
            target = Target.objects.get(name=target_name)
            for alias in aliases[target_name].split(','):
                self.assertTrue(TargetName.objects.filter(target=target, name=alias).exists())


class TestTargetExport(TestCase):
    """
    The use of a list to handle the map returned by StreamingHttpResponse.streaming_content is taken directly from
    the Django httpwrappers tests, as seen here:
    https://github.com/django/django/blob/00ff7a44dee91be59a64559cadeeba0f7386fd6f/tests/httpwrappers/tests.py#L569

    Tests are included for targets with and without aliases due to the previous presence of a bug relating to aliases
    and target export: https://github.com/TOMToolkit/tom_base/issues/265
    """
    def setUp(self):
        self.st = SiderealTargetFactory.create(name='M42', ra=83.8221, dec=-5.3911)
        self.st2 = SiderealTargetFactory.create(name='M52', ra=351.2, dec=61.593)

        self.user = User.objects.create(username='testuser')
        self.client.force_login(self.user)
        assign_perm('tom_targets.view_target',  self.user, self.st)
        assign_perm('tom_targets.view_target',  self.user, self.st2)

    def test_export_all_targets_no_aliases(self):
        response = self.client.get(reverse('targets:export'))
        content = ''.join(line.decode('utf-8') for line in list(response.streaming_content))
        self.assertIn('M42', content)
        self.assertIn('M52', content)

    def test_export_filtered_targets_no_aliases(self):
        response = self.client.get(reverse('targets:export') + '?name=M42')
        content = ''.join(line.decode('utf-8') for line in list(response.streaming_content))
        self.assertIn('M42', content)
        self.assertNotIn('M52', content)

    def test_export_all_targets_with_aliases(self):
        TargetNameFactory.create(name='Messier 42', target=self.st)
        response = self.client.get(reverse('targets:export'))
        content = ''.join(line.decode('utf-8') for line in list(response.streaming_content))
        self.assertIn('M42', content)
        self.assertIn('M52', content)

    def test_export_filtered_targets_with_aliases(self):
        TargetNameFactory.create(name='Messier 42', target=self.st)
        response = self.client.get(reverse('targets:export') + '?name=M42')
        content = ''.join(line.decode('utf-8') for line in list(response.streaming_content))
        self.assertIn('M42', content)
        self.assertNotIn('M52', content)


class TestTargetSearch(TestCase):
    def setUp(self):
        self.st = SiderealTargetFactory.create(name='1337target', ra=269.9983, dec=-29.0698)
        self.st_name = TargetNameFactory.create(name='M42', target=self.st)
        self.st_name2 = TargetNameFactory.create(name='Messier 42', target=self.st)

        self.target2 = SiderealTargetFactory.create(name='Target1309', ra=266.9360, dec=-35.7749)
        self.target2_name = TargetNameFactory.create(name='NGC1309', target=self.target2)
        self.target2_name2 = TargetNameFactory.create(name='PGC 012626', target=self.target2)

        self.user = User.objects.create(username='testuser')
        self.client.force_login(self.user)
        assign_perm('tom_targets.view_target', self.user, self.st)

    def test_search_name_no_results(self):
        response = self.client.get(reverse('targets:list') + '?name=noresults')
        self.assertNotContains(response, '1337target')

    def test_search_name(self):
        response = self.client.get(reverse('targets:list') + '?name=M42')
        self.assertContains(response, '1337target')
        self.assertNotContains(response, '1309Target')

        response = self.client.get(reverse('targets:list') + '?name=Messier 42')
        self.assertContains(response, '1337target')
        self.assertNotContains(response, '1309Target')

    @override_settings(EXTRA_FIELDS=[{'name': 'color', 'type': 'string'}])
    def test_search_extra_fields(self):
        TargetExtra.objects.create(target=self.st, key='color', value='red')

        response = self.client.get(reverse('targets:list') + '?color=red')
        self.assertContains(response, '1337target')

        response = self.client.get(reverse('targets:list') + '?color=blue')
        self.assertNotContains(response, '1337target')

    @override_settings(EXTRA_FIELDS=[{'name': 'birthday', 'type': 'datetime'}])
    def test_search_extra_datetime(self):
        TargetExtra.objects.create(target=self.st, key='birthday', value='2019-02-14')

        response = self.client.get(reverse('targets:list') + '?birthday_after=2019-02-13&birthday_before=2019-02-15')
        self.assertContains(response, '1337target')

    @override_settings(EXTRA_FIELDS=[{'name': 'checked', 'type': 'boolean'}])
    def test_search_extra_boolean(self):
        TargetExtra.objects.create(target=self.st, key='checked', value=False)

        response = self.client.get(reverse('targets:list') + '?checked=3')
        self.assertContains(response, '1337target')

    def test_cone_search_coordinates(self):
        response = self.client.get(reverse('targets:list') + '?cone_search=269.75891,-29.179583,0.25')
        self.assertContains(response, '1337target')
        self.assertNotContains(response, 'Target1309')

    def test_cone_search_target(self):
        response = self.client.get(reverse('targets:list') + '?target_cone_search=1337target,1')
        self.assertContains(response, '1337target')
        self.assertNotContains(response, 'Target1309')


class TestTargetGrouping(TestCase):
    def setUp(self):
        user = User.objects.create(username='testuser')
        self.client.force_login(user)

    def test_view_groupings(self):
        # create a group, check it is added to DB
        group = TargetList(name="test_group")
        group.save()
        self.assertTrue(TargetList.objects.filter(name="test_group").exists())

        # give this user the permission to view it
        user = User.objects.get(username='testuser')
        assign_perm('tom_targets.view_targetlist', user, group)

        response = self.client.get(reverse('targets:targetgrouping'), follow=True)
        self.assertContains(response, group.name)

    def test_create_group(self):
        group_data = {
            'name': 'test_group'
        }
        response = self.client.post(reverse('targets:create-group'), data=group_data)

        self.assertRedirects(response, reverse('targets:targetgrouping'), status_code=302)
        self.assertTrue(TargetList.objects.filter(name=group_data['name']).exists())

    def test_delete_group(self):
        # create a group, check it is added to DB
        group = TargetList(name="test_group")
        group.save()
        self.assertTrue(TargetList.objects.filter(name="test_group").exists())

        # give user permission to delete
        user = User.objects.get(username='testuser')
        assign_perm('tom_targets.delete_targetlist', user, group)

        response = self.client.post(reverse('targets:delete-group', args=(group.pk,)), follow=True)
        self.assertRedirects(response, reverse('targets:targetgrouping'), status_code=302)
        self.assertFalse(TargetList.objects.filter(name='test_group').exists())


class TestTargetAddRemoveGrouping(TestCase):
    def setUp(self):
        user = User.objects.create(username='testuser')
        self.client.force_login(user)
        # create targets
        self.fake_targets = []
        for _ in range(3):
            ft = SiderealTargetFactory.create()
            self.fake_targets.append(ft)
            assign_perm('tom_targets.view_target', user, ft)
            assign_perm('tom_targets.change_target', user, ft)
        # create grouping
        self.fake_grouping = TargetGroupingFactory.create()
        assign_perm('tom_targets.view_targetlist', user, self.fake_grouping)
        # add target[0] to grouping
        self.fake_grouping.targets.add(self.fake_targets[0])

    # Add target[0] and [1] to grouping; [0] already exists and [1] new
    def test_add_selected_to_grouping(self):
        data = {
            'grouping': self.fake_grouping.id,
            'add': True,
            'isSelectAll': 'False',
            'selected-target': [self.fake_targets[0].id, self.fake_targets[1].id],
            'query_string': '',
        }
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)

        self.assertEqual(self.fake_grouping.targets.count(), 2)
        self.assertTrue(self.fake_targets[0] in self.fake_grouping.targets.all())
        self.assertTrue(self.fake_targets[1] in self.fake_grouping.targets.all())

        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertIn(('1 target(s) successfully added to group \'{}\'.'.format(self.fake_grouping.name),
                       SUCCESS), messages)
        self.assertIn(('1 target(s) already in group \'{}\': {}'.format(
            self.fake_grouping.name, self.fake_targets[0].name), WARNING), messages)

    def test_add_to_invalid_grouping(self):
        data = {
            'grouping': -1,
            'add': True,
            'isSelectAll': 'False',
            'selected-target': self.fake_targets[1].id,
            'query_string': '',
        }
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 1)
        self.assertTrue(self.fake_targets[0] in self.fake_grouping.targets.all())
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertTrue('Cannot find the target group with id=-1; ' in messages[0][0])

    # Remove target[0] and [1] from grouping;
    def test_remove_selected_from_grouping(self):
        data = {
            'grouping': self.fake_grouping.id,
            'remove': True,
            'isSelectAll': 'False',
            'selected-target': [self.fake_targets[0].id, self.fake_targets[1].id],
            'query_string': '',
        }
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 0)
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertIn(('1 target(s) successfully removed from group \'{}\'.'.format(self.fake_grouping.name),
                       SUCCESS), messages)
        self.assertIn(('1 target(s) not in group \'{}\': {}'.format(self.fake_grouping.name, self.fake_targets[1].name),
                       WARNING), messages)

    # Add target[0] and [1] to grouping; [0] already exists and [1] new
    def test_move_selected_to_grouping(self):
        data = {
            'grouping': self.fake_grouping.id,
            'move': True,
            'isSelectAll': 'False',
            'selected-target': [self.fake_targets[0].id, self.fake_targets[1].id],
            'query_string': '',
        }
        first_grouping = TargetGroupingFactory.create()
        self.fake_targets[0].targetlist_set.add(first_grouping)
        self.fake_targets[1].targetlist_set.add(first_grouping)
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)

        self.assertEqual(self.fake_grouping.targets.count(), 2)
        self.assertTrue(self.fake_targets[0] in self.fake_grouping.targets.all())
        self.assertTrue(self.fake_targets[1] in self.fake_grouping.targets.all())
        self.assertTrue(self.fake_targets[0] in first_grouping.targets.all())
        self.assertFalse(self.fake_targets[1] in first_grouping.targets.all())

        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertIn(('1 target(s) successfully moved to group \'{}\'.'.format(self.fake_grouping.name),
                       SUCCESS), messages)
        self.assertIn(('1 target(s) already in group \'{}\': {}'.format(
            self.fake_grouping.name, self.fake_targets[0].name), WARNING), messages)

    def test_move_all_to_grouping_filtered_by_sidereal(self):
        data = {
            'grouping': self.fake_grouping.id,
            'move': True,
            'isSelectAll': 'True',
            'selected-target': [],
            'query_string': 'type=SIDEREAL&name=&key=&value=&targetlist__name=',
        }
        first_grouping = TargetGroupingFactory.create()
        first_grouping.targets.add(*self.fake_targets)
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 3)
        self.assertEqual(first_grouping.targets.count(), 1)
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertIn(('2 target(s) successfully moved to group \'{}\'.'.format(self.fake_grouping.name),
                       SUCCESS), messages)
        self.assertIn((
            '1 target(s) already in group \'{}\': {}'.format(self.fake_grouping.name, self.fake_targets[0].name),
            WARNING), messages
        )

    def test_empty_data(self):
        self.client.post(reverse('targets:add-remove-grouping'), data={'query_string': ''})
        self.assertEqual(self.fake_grouping.targets.count(), 1)

    def test_permission_denied(self):
        new_user = User.objects.create(username='newuser')
        self.client.force_login(new_user)
        data = {
            'grouping': self.fake_grouping.id,
            'add': True,
            'isSelectAll': 'False',
            'selected-target': [self.fake_targets[0].id, self.fake_targets[1].id],
            'query_string': '',
        }
        self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 1)

    def test_add_all_to_grouping_filtered_by_sidereal(self):
        data = {
            'grouping': self.fake_grouping.id,
            'add': True,
            'isSelectAll': 'True',
            'selected-target': [],
            'query_string': 'type=SIDEREAL&name=&key=&value=&targetlist__name=',
        }
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 3)
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertIn(('2 target(s) successfully added to group \'{}\'.'.format(self.fake_grouping.name),
                       SUCCESS), messages)
        self.assertIn((
            '1 target(s) already in group \'{}\': {}'.format(self.fake_grouping.name, self.fake_targets[0].name),
            WARNING), messages
        )

    def test_remove_all_from_grouping_filtered_by_sidereal(self):
        data = {
            'grouping': self.fake_grouping.id,
            'remove': True,
            'isSelectAll': 'True',
            'selected-target': [],
            'query_string': 'type=SIDEREAL&name=&key=&value=&targetlist__name=',
        }
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 0)
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertIn(('1 target(s) successfully removed from group \'{}\'.'.format(self.fake_grouping.name),
                       SUCCESS), messages)
        self.assertIn(('2 target(s) not in group \'{}\': {}'.format(
            self.fake_grouping.name, self.fake_targets[1].name + ', ' + self.fake_targets[2].name
        ), WARNING), messages)

    def test_remove_all_from_grouping_filtered_by_grouping(self):
        data = {
            'grouping': self.fake_grouping.id,
            'remove': True,
            'isSelectAll': 'True',
            'selected-target': [],
            'query_string': 'type=&name=&key=&value=&targetlist__name=' + str(self.fake_grouping.id),
        }
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 0)
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertIn(('1 target(s) successfully removed from group \'{}\'.'.format(self.fake_grouping.name),
                       SUCCESS), messages)

    def test_add_remove_with_random_query_string(self):
        data = {
            'grouping': self.fake_grouping.id,
            'remove': True,
            'isSelectAll': 'True',
            'selected-target': [],
            'query_string': 'asdfghjk@!#$%6',
        }
        self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 0)

    def test_add_remove_from_grouping_empty_query_string(self):
        data = {
            'grouping': self.fake_grouping.id,
            'remove': True,
            'isSelectAll': 'True',
            'selected-target': [],
        }
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data)
        self.assertEqual(self.fake_grouping.targets.count(), 0)
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertIn(('1 target(s) successfully removed from group \'{}\'.'.format(self.fake_grouping.name),
                       SUCCESS), messages)
        self.assertIn(('2 target(s) not in group \'{}\': {}'.format(
            self.fake_grouping.name, self.fake_targets[1].name + ', ' + self.fake_targets[2].name
        ), WARNING), messages)

    def test_persist_filter(self):
        data = {'query_string': 'type=SIDEREAL&name=B&key=C&value=123&targetlist__name=1'}
        expected_query_dict = {
            'type': 'SIDEREAL',
            'name': 'B',
            'key': 'C',
            'value': '123',
            'targetlist__name': '1'}
        response = self.client.post(reverse('targets:add-remove-grouping'), data=data, follow=True)
        response_query_dict = response.context['filter'].data.dict()
        self.assertEqual(response_query_dict, expected_query_dict)

    def test_persist_filter_empty(self):
        expected_query_dict = {}
        response = self.client.post(reverse('targets:add-remove-grouping'), data={}, follow=True)
        response_query_dict = response.context['filter'].data
        self.assertEqual(response_query_dict, expected_query_dict)



from django.contrib import admin

from tom_observations.models import DynamicCadence, ObservationGroup, ObservationRecord


class DynamicCadenceAdmin(admin.ModelAdmin):
    model = DynamicCadence


class ObservationGroupAdmin(admin.ModelAdmin):
    model = ObservationGroup


class ObservationRecordAdmin(admin.ModelAdmin):
    model = ObservationRecord


admin.site.register(DynamicCadence, DynamicCadenceAdmin)
admin.site.register(ObservationGroup, ObservationGroupAdmin)
admin.site.register(ObservationRecord, ObservationRecordAdmin)



from django.conf import settings
from django_filters import rest_framework as drf_filters
from guardian.mixins import PermissionListMixin
from guardian.shortcuts import assign_perm, get_objects_for_user
from rest_framework import status
from rest_framework.mixins import CreateModelMixin, DestroyModelMixin, ListModelMixin
from rest_framework.parsers import MultiPartParser
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from tom_common.hooks import run_hook
from tom_dataproducts.data_processor import run_data_processor
from tom_dataproducts.filters import DataProductFilter
from tom_dataproducts.models import DataProduct, ReducedDatum
from tom_dataproducts.serializers import DataProductSerializer


class DataProductViewSet(CreateModelMixin, DestroyModelMixin, ListModelMixin, GenericViewSet, PermissionListMixin):
    """
    Viewset for DataProduct objects. Supports list, create, and delete.

    To view supported query parameters, please use the OPTIONS endpoint, which can be accessed through the web UI.

    **Please note that ``groups`` are an accepted query parameters for the ``CREATE`` endpoint. The groups parameter
    will specify which ``groups`` can view the created ``DataProduct``. If no ``groups`` are specified, the
    ``DataProduct`` will only be visible to the user that created the ``DataProduct``. Make sure to check your
    ``groups``!!**
    """
    queryset = DataProduct.objects.all()
    serializer_class = DataProductSerializer
    filter_backends = (drf_filters.DjangoFilterBackend,)
    filterset_class = DataProductFilter
    permission_required = 'tom_dataproducts.view_dataproduct'
    parser_classes = [MultiPartParser]

    def create(self, request, *args, **kwargs):
        request.data['data'] = request.FILES['file']
        response = super().create(request, *args, **kwargs)

        if response.status_code == status.HTTP_201_CREATED:
            dp = DataProduct.objects.get(pk=response.data['id'])
            try:
                run_hook('data_product_post_upload', dp)
                reduced_data = run_data_processor(dp)
                if not settings.TARGET_PERMISSIONS_ONLY:
                    for group in response.data['group']:
                        assign_perm('tom_dataproducts.view_dataproduct', group, dp)
                        assign_perm('tom_dataproducts.delete_dataproduct', group, dp)
                        assign_perm('tom_dataproducts.view_reduceddatum', group, reduced_data)
            except Exception:
                ReducedDatum.objects.filter(data_product=dp).delete()
                dp.delete()
                return Response({'Data processing error': '''There was an error in processing your DataProduct into \
                                                             individual ReducedDatum objects.'''},
                                status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        return response

    def get_queryset(self):
        """
        Gets the set of ``DataProduct`` objects that the user has permission to view.

        :returns: Set of ``DataProduct`` objects
        :rtype: QuerySet
        """
        if settings.TARGET_PERMISSIONS_ONLY:
            return super().get_queryset().filter(
                target__in=get_objects_for_user(self.request.user, 'tom_targets.view_target')
            )
        else:
            return get_objects_for_user(self.request.user, 'tom_dataproducts.view_dataproduct')



from django import forms
from django.contrib.auth.forms import UserCreationForm, UsernameField
from django.contrib.auth.models import User, Group


class ChangeUserPasswordForm(forms.Form):
    password = forms.CharField(widget=forms.PasswordInput())


class GroupForm(forms.ModelForm):
    users = forms.ModelMultipleChoiceField(User.objects.all(), required=False, widget=forms.CheckboxSelectMultiple)

    class Meta:
        model = Group
        fields = ('name', 'users')

    def save(self, *args, **kwargs):
        instance = super().save(*args, **kwargs)
        instance.user_set.set(self.cleaned_data['users'])
        instance.save()
        return instance


class CustomUserCreationForm(UserCreationForm):
    """
    Form used for creation of new users and update of existing users.
    """
    email = forms.EmailField(required=True)
    groups = forms.ModelMultipleChoiceField(Group.objects.all().exclude(name='Public'),
                                            required=False, widget=forms.CheckboxSelectMultiple)

    class Meta:
        model = User
        fields = ('username', 'first_name', 'last_name', 'email', 'groups')
        field_classes = {'username': UsernameField}

    def save(self, commit=True):
        user = super(forms.ModelForm, self).save(commit=False)
        # Because this form is used for both create and update user, and the user can be updated without modifying the
        # password, we check if the password field has been populated in order to set a new one.
        if self.cleaned_data['password1']:
            user.set_password(self.cleaned_data["password1"])
        if commit:
            user.save()
            self.save_m2m()

        return user



from django import template
from django.db.models import IntegerField
from django.db.models.functions import Cast
from django.conf import settings
from django_comments.models import Comment
from guardian.shortcuts import get_objects_for_user

register = template.Library()


@register.inclusion_tag('tom_common/partials/navbar_login.html', takes_context=True)
def navbar_login(context):
    """
    Renders the username as a link to the user page, as well as the login button. Can be overridden to render additional
    account-related buttons.
    """
    return {'user': context['user']}


@register.simple_tag
def comments_enabled():
    """
    Returns the TOM setting specifying whether or not comments are enabled
    """
    try:
        return settings.COMMENTS_ENABLED
    except AttributeError:
        return True


@register.simple_tag
def verbose_name(instance, field_name):
    """
    Displays the more descriptive field name from a Django model field
    """
    return instance._meta.get_field(field_name).verbose_name.title()


@register.inclusion_tag('comments/list.html', takes_context=True)
def recent_comments(context, limit=10):
    """
    Displays a list of the most recent comments in the TOM up to the given limit, or 10 if not specified.

    Comments will only be displayed for targets which the logged-in user has permission to view.
    """
    user = context['request'].user
    targets_for_user = get_objects_for_user(user, 'tom_targets.view_target')

    # In django-contrib-comments, the Comment model has a field ``object_pk`` which refers to the primary key
    # of the object it is related to, i.e., a comment on a ``Target`` has an ``object_pk`` corresponding with the
    # ``Target`` id. The ``object_pk`` is stored as a TextField.

    # In order to filter on ``object_pk`` with an iterable of ``IntegerFields`` using the ``in`` comparator,
    # we have to cast the ``object_pk`` to an int and annotate it as ``object_pk_as_int``.
    return {
        'comment_list': Comment.objects.annotate(
            object_pk_as_int=Cast('object_pk', output_field=IntegerField())
        ).filter(
            object_pk_as_int__in=targets_for_user
        ).order_by('-submit_date')[:limit]
    }


@register.filter
def truncate_number(value):
    """
    Truncates a numerical value to four decimal places for display purposes. Etienne Bachelet advised that three
    decimal places was insufficient precision, but that four would be more acceptable.
    """
    try:
        return '%.4f' % value
    except Exception:
        return value


@register.simple_tag
def tom_name():
    return getattr(settings, 'TOM_NAME', 'TOM Toolkit')



from django.conf import settings
from django.contrib.auth.models import Group
from django.db.models.query import QuerySet
from guardian.shortcuts import assign_perm, get_objects_for_user
from rest_framework import serializers

from tom_common.serializers import GroupSerializer
from tom_observations.models import ObservationGroup, ObservationRecord


class ObservationGroupField(serializers.RelatedField):
    """
    ``RelatedField`` used to display ObservationGroups and DynamicCadences in a prettier fashion than offered by default
    DRF implementations.
    """
    def to_representation(self, instance: ObservationGroup) -> dict:
        return {
            'name': instance.name,
            'dynamic_cadences': [dc.__str__() for dc in instance.dynamiccadence_set.all()]
        }


class ObservationRecordSerializer(serializers.ModelSerializer):
    groups = GroupSerializer(many=True, required=False)  # TODO: return groups in detail and list
    observation_groups = ObservationGroupField(many=True, read_only=True, source='observationgroup_set')
    status = serializers.CharField(required=False)

    class Meta:
        model = ObservationRecord
        fields = '__all__'

    def create(self, validated_data):
        groups = validated_data.pop('groups', [])

        obsr = ObservationRecord.objects.create(**validated_data)

        group_serializer = GroupSerializer(data=groups, many=True)
        if group_serializer.is_valid() and settings.TARGET_PERMISSIONS_ONLY is False:
            for group in groups:
                group_instance = Group.objects.get(pk=group['id'])
                assign_perm('tom_observations.view_observationrecord', group_instance, obsr)
                assign_perm('tom_observations.change_observationrecord', group_instance, obsr)
                assign_perm('tom_observations.delete_observationrecord', group_instance, obsr)

        return obsr


class ObservationRecordFilteredPrimaryKeyRelatedField(serializers.PrimaryKeyRelatedField):
    # This PrimaryKeyRelatedField subclass is used to implement get_queryset based on the permissions of the user
    # submitting the request. The pattern was taken from this StackOverflow answer: https://stackoverflow.com/a/32683066

    def get_queryset(self) -> QuerySet:
        request = self.context.get('request', None)
        queryset = super().get_queryset()
        if not (request and queryset):
            return None
        if settings.TARGET_PERMISSIONS_ONLY:
            return ObservationRecord.objects.all()
        else:
            return get_objects_for_user(request.user, 'tom_observations.change_observation')



import json
from requests import Response
from unittest.mock import patch

from django.test import TestCase

from tom_common.exceptions import ImproperCredentialsException
from tom_observations.facilities.soar import make_request, SOARBaseObservationForm, SOARImagingObservationForm
from tom_observations.facilities.soar import SOARSpectroscopyObservationForm
from tom_observations.tests.factories import NonSiderealTargetFactory, SiderealTargetFactory


instrument_response = {
    '2M0-FLOYDS-SCICAM': {
        'type': 'SPECTRA', 'class': '2m0', 'name': '2.0 meter FLOYDS', 'optical_elements': {
            'slits': [
              {'name': '6.0 arcsec slit', 'code': 'slit_6.0as', 'schedulable': True, 'default': False},
              {'name': '1.6 arcsec slit', 'code': 'slit_1.6as', 'schedulable': True, 'default': False},
              {'name': '2.0 arcsec slit', 'code': 'slit_2.0as', 'schedulable': True, 'default': False},
              {'name': '1.2 arcsec slit', 'code': 'slit_1.2as', 'schedulable': True, 'default': False}
            ]
        }
    },
    '0M4-SCICAM-SBIG': {
        'type': 'IMAGE', 'class': '0m4', 'name': '0.4 meter SBIG', 'optical_elements': {
            'filters': [
                {'name': 'Opaque', 'code': 'opaque', 'schedulable': False, 'default': False},
                {'name': '100um Pinhole', 'code': '100um-Pinhole', 'schedulable': False, 'default': False},
            ]
        },
    },
    'SOAR_GHTS_REDCAM': {
        'type': 'SPECTRA', 'class': '4m0', 'name': 'Goodman Spectrograph RedCam', 'optical_elements': {
            'gratings': [
                {'name': '400 line grating', 'code': 'SYZY_400', 'schedulable': True, 'default': True},
            ],
            'slits': [
                {'name': '1.0 arcsec slit', 'code': 'slit_1.0as', 'schedulable': True, 'default': True}
            ]
        },
    },
    'SOAR_GHTS_REDCAM_IMAGER': {
        'type': 'IMAGE', 'class': '4m0', 'name': 'Goodman Spectrograph RedCam Imager', 'optical_elements': {
            'filters': [
                {'name': 'Clear', 'code': 'air', 'schedulable': True, 'default': False},
                {'name': 'GHTS u-SDSS', 'code': 'u-SDSS', 'schedulable': False, 'default': False},
                {'name': 'GHTS g-SDSS', 'code': 'g-SDSS', 'schedulable': True, 'default': False},
                {'name': 'GHTS r-SDSS', 'code': 'r-SDSS', 'schedulable': True, 'default': True},
                {'name': 'GHTS i-SDSS', 'code': 'i-SDSS', 'schedulable': True, 'default': False},
                {'name': 'GHTS z-SDSS', 'code': 'z-SDSS', 'schedulable': False, 'default': False},
                {'name': 'GHTS VR', 'code': 'VR', 'schedulable': True, 'default': False}
            ]
        },
    }
}


class TestMakeRequest(TestCase):

    @patch('tom_observations.facilities.soar.requests.request')
    def test_make_request(self, mock_request):
        mock_response = Response()
        mock_response._content = str.encode(json.dumps({'test': 'test'}))
        mock_response.status_code = 200
        mock_request.return_value = mock_response

        self.assertDictEqual({'test': 'test'}, make_request('GET', 'google.com', headers={'test': 'test'}).json())

        mock_response.status_code = 403
        mock_request.return_value = mock_response
        with self.assertRaises(ImproperCredentialsException):
            make_request('GET', 'google.com', headers={'test': 'test'})


class TestSOARBaseObservationForm(TestCase):

    def setUp(self):
        self.st = SiderealTargetFactory.create()
        self.nst = NonSiderealTargetFactory.create(scheme='MPC_MINOR_PLANET')
        self.valid_form_data = {
            'name': 'test', 'facility': 'SOAR', 'target_id': self.st.id, 'ipp_value': 0.5, 'start': '2020-11-03',
            'end': '2020-11-04', 'exposure_count': 1, 'exposure_time': 30, 'max_airmass': 3,
            'min_lunar_distance': 20, 'observation_mode': 'NORMAL',
            'proposal': 'sampleproposal', 'filter': 'opaque', 'instrument_type': 'SOAR_GHTS_REDCAM_IMAGER'
        }
        self.instrument_choices = [(k, v['name']) for k, v in instrument_response.items() if 'SOAR' in k]
        self.filter_choices = set([
            (f['code'], f['name']) for ins in instrument_response.values() for f in
            ins['optical_elements'].get('filters', []) + ins['optical_elements'].get('slits', [])
        ])
        self.proposal_choices = [('sampleproposal', 'Sample Proposal')]

    @patch('tom_observations.facilities.soar.make_request')
    @patch('tom_observations.facilities.soar.cache')
    def test_get_instruments(self, mock_cache, mock_make_request):
        mock_response = Response()
        mock_response._content = str.encode(json.dumps(instrument_response))
        mock_response.status_code = 200
        mock_make_request.return_value = mock_response

        # Test that cached value is returned
        with self.subTest():
            test_instruments = {'test instrument': {'type': 'IMAGE'}}
            mock_cache.get.return_value = test_instruments

            instruments = SOARBaseObservationForm._get_instruments()
            self.assertDictContainsSubset({'test instrument': {'type': 'IMAGE'}}, instruments)
            self.assertNotIn('0M4-SCICAM-SBIG', instruments)

        # Test that empty cache results in mock_instruments, and cache.set is called
        with self.subTest():
            mock_cache.get.return_value = None

            instruments = SOARBaseObservationForm._get_instruments()
            self.assertIn('SOAR_GHTS_REDCAM_IMAGER', instruments)
            self.assertDictContainsSubset({'type': 'IMAGE'}, instruments['SOAR_GHTS_REDCAM_IMAGER'])
            self.assertNotIn('0M4-SCICAM-SBIG', instruments)
            mock_cache.set.assert_called()

    @patch('tom_observations.facilities.soar.SOARBaseObservationForm.proposal_choices')
    @patch('tom_observations.facilities.soar.SOARBaseObservationForm.filter_choices')
    @patch('tom_observations.facilities.soar.SOARBaseObservationForm.instrument_choices')
    @patch('tom_observations.facilities.soar.SOARBaseObservationForm.validate_at_facility')
    def test_instrument_to_type(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test instrument_to_type method."""
        self.assertEqual('EXPOSE', SOARBaseObservationForm.instrument_to_type('SOAR_GHTS_REDCAM_IMAGER'))
        self.assertEqual('SPECTRUM', SOARBaseObservationForm.instrument_to_type('SOAR_GHTS_REDCAM'))


@patch('tom_observations.facilities.soar.SOARImagingObservationForm._get_instruments')
class TestSOARImagingObservationForm(TestCase):
    def test_instrument_choices(self, mock_get_instruments):
        """Test SOARImagingObservationForm._instrument_choices."""
        mock_get_instruments.return_value = {k: v for k, v in instrument_response.items() if 'SOAR' in k}

        inst_choices = SOARImagingObservationForm.instrument_choices()
        self.assertIn(('SOAR_GHTS_REDCAM_IMAGER', 'Goodman Spectrograph RedCam Imager'), inst_choices)
        self.assertNotIn(('SOAR_GHTS_REDCAM', 'Goodman Spectrograph RedCam'), inst_choices)
        self.assertEqual(len(inst_choices), 1)

    def test_filter_choices(self, mock_get_instruments):
        """Test SOARImagingObservationForm._filter_choices."""
        mock_get_instruments.return_value = {k: v for k, v in instrument_response.items() if 'SOAR' in k}

        filter_choices = SOARImagingObservationForm.filter_choices()
        for expected in [('air', 'Clear'), ('u-SDSS', 'GHTS u-SDSS'), ('g-SDSS', 'GHTS g-SDSS'),
                         ('r-SDSS', 'GHTS r-SDSS'), ('i-SDSS', 'GHTS i-SDSS'), ('z-SDSS', 'GHTS z-SDSS'),
                         ('VR', 'GHTS VR')]:
            self.assertIn(expected, filter_choices)
        for not_expected in [('slit_1.0as', '1.0 arcsec slit')]:
            self.assertNotIn(not_expected, filter_choices)
        self.assertEqual(len(filter_choices), 7)


class TestSOARSpectroscopyObservationForm(TestCase):

    @patch('tom_observations.facilities.soar.SOARSpectroscopyObservationForm._get_instruments')
    def test_instrument_choices(self, mock_get_instruments):
        """Test SOARSpectroscopyObservationForm._instrument_choices."""
        mock_get_instruments.return_value = {k: v for k, v in instrument_response.items() if 'SOAR' in k}

        inst_choices = SOARSpectroscopyObservationForm.instrument_choices()
        self.assertIn(('SOAR_GHTS_REDCAM', 'Goodman Spectrograph RedCam'), inst_choices)
        self.assertNotIn(('SOAR_GHTS_REDCAM_IMAGER', 'Goodman Spectrograph RedCam Imager'), inst_choices)
        self.assertEqual(len(inst_choices), 1)

    @patch('tom_observations.facilities.soar.SOARSpectroscopyObservationForm._get_instruments')
    def test_filter_choices(self, mock_get_instruments):
        """Test SOARSpectroscopyObservationForm._filter_choices."""
        mock_get_instruments.return_value = {k: v for k, v in instrument_response.items() if 'SOAR' in k}

        filter_choices = SOARSpectroscopyObservationForm.filter_choices()
        for expected in [('slit_1.0as', '1.0 arcsec slit')]:
            self.assertIn(expected, filter_choices)
        for not_expected in [('u-SDSS', 'GHTS u-SDSS'), ('i-SDSS', 'GHTS i-SDSS')]:
            self.assertNotIn(not_expected, filter_choices)
        self.assertEqual(len(filter_choices), 1)

    @patch('tom_observations.facilities.soar.SOARSpectroscopyObservationForm.proposal_choices')
    @patch('tom_observations.facilities.soar.SOARSpectroscopyObservationForm.filter_choices')
    @patch('tom_observations.facilities.soar.SOARSpectroscopyObservationForm.instrument_choices')
    @patch('tom_observations.facilities.soar.SOARSpectroscopyObservationForm.validate_at_facility')
    def test_build_instrument_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_validate.return_value = []
        mock_insts.return_value = [(k, v['name']) for k, v in instrument_response.items() if 'SPECTRA' in v['type']]
        mock_filters.return_value = set([
            (f['code'], f['name']) for ins in instrument_response.values() for f in
            ins['optical_elements'].get('slits', [])
            ] + [('None', 'None')])
        mock_proposals.return_value = [('sampleproposal', 'Sample Proposal')]

        st = SiderealTargetFactory.create()
        valid_form_data = {
            'name': 'test', 'facility': 'SOAR', 'target_id': st.id, 'ipp_value': 0.5, 'start': '2020-11-03',
            'end': '2020-11-04', 'exposure_count': 1, 'exposure_time': 30.0, 'max_airmass': 3,
            'min_lunar_distance': 20, 'observation_mode': 'NORMAL', 'proposal': 'sampleproposal',
            'filter': 'slit_1.0as', 'instrument_type': 'SOAR_GHTS_REDCAM', 'rotator_angle': 1.0
        }

        # Test that optical_elements['slit'] and optical_elements['grating] are populated when filter is included
        with self.subTest():
            form = SOARSpectroscopyObservationForm(valid_form_data)
            self.assertTrue(form.is_valid())
            self.assertEqual(
                [{'exposure_count': valid_form_data['exposure_count'],
                  'exposure_time': valid_form_data['exposure_time'],
                  'optical_elements': {'slit': valid_form_data['filter'], 'grating': 'SYZY_400'},
                  'rotator_mode': 'SKY',
                  'extra_params': {'rotator_angle': valid_form_data['rotator_angle']}
                  }], form._build_instrument_config()
            )



from django.core.management.base import BaseCommand
import sys
import os
from django.conf import settings
from django.template.loader import get_template
from django.core.management import call_command
from django.core.management.utils import get_random_secret_key
from django.utils import timezone
from django.contrib.auth.models import Group, User

BASE_DIR = settings.BASE_DIR


class Command(BaseCommand):
    help = 'A configuration helper for new TOMs.'
    context = {}

    def exit(self, msg='Exiting...'):
        self.stdout.write(msg)
        sys.exit(0)

    def ok(self):
        self.stdout.write(self.style.SUCCESS('OK'))

    def status(self, msg):
        self.stdout.write(msg, ending='')
        sys.stdout.flush()

    def welcome_banner(self):
        welcome_text = (
            'Welcome to the tom_setup helper script. This will help you get started with a new TOM.\n'
            'DO NOT RUN THIS SCRIPT ON AN EXISTING TOM. It will override any custom settings you may '
            'already have.\n'
        )
        prompt = 'Do you wish to continue? {}'.format(self.style.WARNING('[y/N] '))
        self.stdout.write(welcome_text)
        while True:
            response = input(prompt).lower()
            if not response or response == 'n':
                self.stdout.write('Aborting installation.')
                self.exit()
            elif response == 'y':
                break
            else:
                self.stdout.write('Invalid response. Please try again.')

    def check_python(self):
        self.status('Checking Python version... ')
        major = sys.version_info.major
        minor = sys.version_info.minor
        if major < 3 or minor < 7:
            self.exit('Incompatible Python version found. Please install Python >= 3.7')
        self.ok()

    def create_project_dirs(self):
        self.status('Creating project directories... ')
        try:
            os.mkdir(os.path.join(BASE_DIR, 'data'))
        except FileExistsError:
            pass
        try:
            os.mkdir(os.path.join(BASE_DIR, 'templates'))
        except FileExistsError:
            pass
        try:
            os.mkdir(os.path.join(BASE_DIR, 'static'))
        except FileExistsError:
            pass
        # os.mknod requires superuser permissions on osx, so create a blank file instead
        try:
            open(os.path.join(BASE_DIR, 'static/.keep'), 'w').close()
        except FileExistsError:
            pass
        try:
            os.mkdir(os.path.join(BASE_DIR, 'tmp'))
        except FileExistsError:
            pass
        self.ok()

    def run_migrations(self):
        self.status('Running initial migrations... ')
        call_command('migrate', verbosity=0, interactive=False)
        self.ok()

    def get_target_type(self):
        allowed_types = {
            '1': 'SIDEREAL',
            '2': 'NON_SIDEREAL'
        }
        options_str = ['{}) {}'.format(key, target_type) for key, target_type in allowed_types.items()]
        prompt = 'Which target type will your project use? {} '.format(self.style.WARNING(", ".join(options_str)))
        target_type = input(prompt)
        try:
            self.context['TARGET_TYPE'] = allowed_types[target_type]
        except KeyError:
            self.stdout.write('Error: invalid choice {}'.format(target_type))
            self.get_target_type()

    def get_hint_preference(self):
        help_message_info = (
            'Help messages can be configured to appear to give suggestions on commonly customized functions. If '
            'enabled now, they can be turned off by changing HINTS_ENABLED to False in settings.py.\n'
        )
        prompt = 'Would you like to enable hints? {}'.format(self.style.WARNING('[y/N] '))
        self.stdout.write(help_message_info)
        while True:
            response = input(prompt).lower()
            if not response or response == 'n':
                self.context['HINTS_ENABLED'] = False
            elif response == 'y':
                self.context['HINTS_ENABLED'] = True
            else:
                self.stdout.write('Invalid response. Please try again.')
                continue

            break

    def get_permissions_preference(self):
        permissions_message_info = (
            'The TOM Toolkit permissions system allows you to either restrict access based on which user groups have '
            'access to particular targets, or lets you restrict access for Targets, ObservationRecords, and '
            'DataProducts individually.'
        )
        prompt = 'Would you like to the enable the second, more comprehensive permissions system? {}'.format(
            self.style.WARNING('[y/N] ')
        )
        self.stdout.write(permissions_message_info)
        while True:
            response = input(prompt).lower()
            if not response or response == 'n':
                self.context['TARGET_PERMISSIONS_ONLY'] = True
            elif response == 'y':
                self.context['TARGET_PERMISSIONS_ONLY'] = False
            else:
                self.stdout.write('Invalid response. Please try again.')
                continue

            break

    def generate_secret_key(self):
        self.status('Generating secret key... ')
        self.context['SECRET_KEY'] = get_random_secret_key()
        self.ok()

    def generate_config(self):
        self.status('Generating settings.py... ')
        template = get_template('tom_setup/settings.tmpl')
        rendered = template.render(self.context)

        # TODO: Ugly hack to get project name
        settings_location = os.path.join(BASE_DIR, os.path.basename(BASE_DIR), 'settings.py')
        if not os.path.exists(settings_location):
            msg = 'Could not determine settings.py location. Writing settings.py out to {}. Please copy file to \
                   the proper location after script finishes.'.format(settings_location)
            self.stdout.write(self.style.WARNING(msg))
        with open(settings_location, 'w+') as settings_file:
            settings_file.write(rendered)

        self.ok()

    def generate_urls(self):
        self.status('Generating urls.py... ')
        template = get_template('tom_setup/urls.tmpl')
        rendered = template.render(self.context)

        # TODO: Ugly hack to get project name
        urls_location = os.path.join(BASE_DIR, os.path.basename(BASE_DIR), 'urls.py')
        if not os.path.exists(urls_location):
            msg = 'Could not determine urls.py location. Writing urls.py out to {}. Please copy file to \
                   the proper location after script finishes.'.format(urls_location)
            self.stdout.write(self.style.WARNING(msg))
        with open(urls_location, 'w+') as urls_file:
            urls_file.write(rendered)

        self.ok()

    def create_pi(self):
        self.stdout.write('Please create a Principal Investigator (PI) for your project')
        call_command('createsuperuser')

    def create_public_group(self):
        self.status('Setting up Public group... ')
        group = Group.objects.create(name='Public')
        group.user_set.add(*User.objects.all())
        group.save()
        self.ok()

    def complete(self):
        self.exit(
            self.style.SUCCESS('Setup complete! Run ./manage.py migrate && ./manage.py runserver to start your TOM.')
        )

    def handle(self, *args, **options):
        self.context['CREATE_DATE'] = timezone.now()
        self.context['PROJECT_NAME'] = os.path.basename(BASE_DIR)
        self.welcome_banner()
        self.check_python()
        self.create_project_dirs()
        self.generate_secret_key()
        self.get_target_type()
        self.get_hint_preference()
        self.generate_config()
        self.generate_urls()
        self.run_migrations()
        self.create_pi()
        self.create_public_group()
        self.complete()



# Generated by Django 2.1.4 on 2019-02-14 17:24

import datetime

from django.db import migrations, models
import django.db.models.deletion
import tom_dataproducts.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('tom_observations', '0001_initial'),
        ('tom_targets', '0004_auto_20190123_2010'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataProduct',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('product_id', models.CharField(max_length=255, null=True, unique=True)),
                ('data', models.FileField(default=None, null=True, upload_to=tom_dataproducts.models.data_product_path)),
                ('extra_data', models.TextField(blank=True, default='')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('tag', models.CharField(blank=True, choices=[('photometry', 'Photometry'), ('fits_file', 'Fits File'), ('spectroscopy', 'Spectroscopy')], default='', max_length=50)),
                ('featured', models.BooleanField(default=False)),
            ],
            options={
                'ordering': ('-created',),
                'get_latest_by': ('modified',),
            },
        ),
        migrations.CreateModel(
            name='DataProductGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
            options={
                'ordering': ('-created',),
            },
        ),
        migrations.CreateModel(
            name='ReducedDatum',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('data_type', models.CharField(choices=[('SPECTROSCOPY', 'Spectroscopy'), ('PHOTOMETRY', 'Photometry')], default='', max_length=100)),
                ('source_name', models.CharField(default='', max_length=100)),
                ('source_location', models.CharField(default='', max_length=200)),
                ('timestamp', models.DateTimeField(db_index=True, default=datetime.datetime.now)),
                ('value', models.TextField()),
                ('data_product', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='tom_dataproducts.DataProduct')),
                ('target', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tom_targets.Target')),
            ],
        ),
        migrations.AddField(
            model_name='dataproduct',
            name='group',
            field=models.ManyToManyField(to='tom_dataproducts.DataProductGroup'),
        ),
        migrations.AddField(
            model_name='dataproduct',
            name='observation_record',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to='tom_observations.ObservationRecord'),
        ),
        migrations.AddField(
            model_name='dataproduct',
            name='target',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tom_targets.Target'),
        ),
    ]



import json

from django.db import migrations, models


def copy_cadence_fields_to_dynamic_cadence(apps, schema_editor):
    observation_groups = apps.get_model('tom_observations', 'ObservationGroup')
    for row in observation_groups.objects.exclude(cadence_strategy=''):
        dynamic_cadence = apps.get_model('tom_observations', 'DynamicCadence')
        try:
            cadence_parameters = json.loads(getattr(row, 'cadence_parameters'))
        except json.decoder.JSONDecodeError:
            cadence_parameters = {}
        new_dynamic_cadence = dynamic_cadence(
            observation_group=row,
            cadence_strategy=getattr(row, 'cadence_strategy'),
            cadence_parameters=cadence_parameters,
            active=True,
            created=getattr(row, 'created'),
            modified=getattr(row, 'modified')
        )
        new_dynamic_cadence.save()


class Migration(migrations.Migration):
    dependencies = [
        ('tom_observations', '0009_observationrecord_user'),
    ]

    operations = [
        migrations.CreateModel(
            name='DynamicCadence',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('cadence_strategy', models.CharField(max_length=100, blank=False, default=None,
                                                      verbose_name='Cadence strategy used for this DynamicCadence')),
                ('cadence_parameters', models.JSONField(blank=False, null=False,
                                                        verbose_name='Cadence-specific parameters')),
                ('active', models.BooleanField(verbose_name='Active',
                                               help_text='''Whether or not this DynamicCadence should continue
                                                          to submit observations.''')),
                ('created', models.DateTimeField(auto_now_add=True,
                                                 help_text='The time which this DynamicCadence was created.')),
                ('modified', models.DateTimeField(auto_now=True,
                                                  help_text='The time which this DynamicCadence was modified.')),
            ]
        ),

        migrations.AddField(
            model_name='dynamiccadence',
            name='observation_group',
            field=models.ForeignKey(on_delete=models.deletion.CASCADE, null=False, default=None,
                                    to='tom_observations.ObservationGroup'),
        ),

        migrations.RunPython(copy_cadence_fields_to_dynamic_cadence, reverse_code=migrations.RunPython.noop),

        migrations.RemoveField(
            model_name='observationgroup',
            name='cadence_strategy'
        ),

        migrations.RemoveField(
            model_name='observationgroup',
            name='cadence_parameters'
        )
    ]


import copy

from django.contrib.auth.models import User, Group
from django.urls import reverse
from guardian.shortcuts import assign_perm, get_objects_for_user
from rest_framework import status
from rest_framework.test import APITestCase

from tom_targets.tests.factories import SiderealTargetFactory, NonSiderealTargetFactory
from tom_targets.tests.factories import TargetExtraFactory, TargetNameFactory
from tom_targets.models import Target, TargetExtra, TargetName


class TestTargetViewset(APITestCase):
    def setUp(self):
        # Create test user with permissions
        self.user = User.objects.create(username='testuser')
        self.st = SiderealTargetFactory.create(name='test target', targetextra_set=None, aliases=None)
        self.st2 = SiderealTargetFactory.create()
        self.nst = NonSiderealTargetFactory.create()
        assign_perm('tom_targets.view_target', self.user, self.st)
        assign_perm('tom_targets.view_target', self.user, self.nst)
        assign_perm('tom_targets.add_target', self.user)
        assign_perm('tom_targets.change_target', self.user, self.st)
        assign_perm('tom_targets.change_target', self.user, self.nst)
        assign_perm('tom_targets.delete_target', self.user, self.st)

        # Create test user with subset of permissions
        self.user2 = User.objects.create(username='testuser2')
        assign_perm('tom_targets.view_target', self.user2, self.st)

        # Login with privileged user
        self.client.force_login(self.user)

    def test_target_list(self):
        response = self.client.get(reverse('api:targets-list'))
        self.assertEqual(response.json()['count'], 2)

        # Ensure that a user without view_target permission on all targets can only retrieve the subset of targets for
        # which they have permission
        self.client.force_login(self.user2)
        response = self.client.get(reverse('api:targets-list'))
        self.assertEqual(response.json()['count'], 1)

    def test_target_detail(self):
        response = self.client.get(reverse('api:targets-detail', args=(self.st.id,)))
        self.assertEqual(response.json()['name'], self.st.name)

        # Ensure that a user without view_target permission cannot access the target
        self.client.force_login(self.user2)
        response = self.client.get(reverse('api:targets-detail', args=(self.nst.id,)))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json()['detail'], 'Not found.')

    def test_target_create(self):
        collaborator = User.objects.create(username='test collaborator')
        group = Group.objects.create(name='bourgeoisie')
        group.user_set.add(self.user)
        group.user_set.add(collaborator)

        target_data = {
            'name': 'test_target_name_wtf',
            'type': Target.SIDEREAL,
            'ra': 123.456,
            'dec': -32.1,
            'groups': [
                {'id': group.id}
            ],
            'targetextra_set': [
                {'key': 'foo', 'value': 5}
            ],
            'aliases': [
                {'name': 'alternative name'}
            ]
        }
        response = self.client.post(reverse('api:targets-list'), data=target_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()['name'], target_data['name'])
        self.assertEqual(response.json()['aliases'][0]['name'], target_data['aliases'][0]['name'])
        self.assertEqual(get_objects_for_user(collaborator, 'tom_targets.view_target').first().name,
                         target_data['name'])  # Test that group permissions are respected

        # TODO: For whatever reason, in django-guardian, authenticated users have permission to create objects,
        # regardless of their row-level permissions. This should be addressed eventually--however, we don't provide a
        # way for PIs to restrict create/update ability, simply target access, so this can be ignored at present.
        #
        # self.client.force_login(self.user2)
        # target_data['name'] = 'test_target_create_bad_permissions'
        # target_data['aliases'] = []
        # response = self.client.post(reverse('api:targets-list'), data=target_data)
        # self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        # Ensure targets can't be created with duplicate aliases
        invalid_target_data = copy.deepcopy(target_data)
        invalid_target_data['name'] = 'invalid_name'
        invalid_target_data['aliases'][0]['name'] = 'invalid_name'
        response = self.client.post(reverse('api:targets-list'), data=invalid_target_data)
        self.assertContains(response,
                            'Alias \'invalid_name\' conflicts with Target name \'invalid_name\'.',
                            status_code=status.HTTP_400_BAD_REQUEST)

    def test_target_create_sidereal_missing_parameters(self):
        target_data = {
            'name': 'test_target_name_wtf',
            'type': Target.SIDEREAL,
            'ra': 123.456,
            'targetextra_set': [
                {'key': 'foo', 'value': 5}
            ],
            'aliases': [
                {'name': 'alternative name'}
            ]
        }
        response = self.client.post(reverse('api:targets-list'), data=target_data)
        self.assertContains(response,
                            'The following fields are required for SIDEREAL targets: [\'dec\']',
                            status_code=status.HTTP_400_BAD_REQUEST)

    def test_target_create_non_sidereal_missing_parameters(self):
        target_data = {
            'name': 'test_target_name_wtf',
            'type': Target.NON_SIDEREAL,
            'epoch_of_elements': 2000,
            'inclination': '0.0005',
            'lng_asc_node': '0.12345',
            'arg_of_perihelion': '57',
            'targetextra_set': [
                {'key': 'foo', 'value': 5}
            ],
            'aliases': [
                {'name': 'alternative name'}
            ]
        }
        response = self.client.post(reverse('api:targets-list'), data=target_data)
        self.assertContains(response,
                            'The following fields are required for NON_SIDEREAL targets: [\'eccentricity\']',
                            status_code=status.HTTP_400_BAD_REQUEST)

    def test_target_update(self):
        collaborator = User.objects.create(username='test collaborator')
        group = Group.objects.create(name='bourgeoisie')
        group.user_set.add(self.user)
        group.user_set.add(collaborator)

        updates = {'ra': 123.456, 'groups': [{'id': group.id}]}
        response = self.client.patch(reverse('api:targets-detail', args=(self.st.id,)), data=updates)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.st.refresh_from_db()
        self.assertEqual(self.st.ra, updates['ra'])
        self.assertEqual(get_objects_for_user(collaborator, 'tom_targets.view_target').first().name,
                         self.st.name)  # Test that group permissions are respected

        self.client.force_login(self.user2)
        updates = {'ra': 654.321}
        response = self.client.patch(reverse('api:targets-detail', args=(self.st.id,)), data=updates)
        self.st.refresh_from_db()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_targetname_update(self):
        # Test both create new alias and update alias
        alias = TargetNameFactory.create(name='alias', target=self.st)
        updates = {
            'aliases': [
                {'id': alias.id, 'name': 'update alias'},
                {'name': 'create alias'}
            ]
        }
        response = self.client.patch(reverse('api:targets-detail', args=(self.st.id,)), data=updates)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.st.refresh_from_db()
        alias.refresh_from_db()
        self.assertEqual(self.st.aliases.count(), 2)
        self.assertEqual(alias.name, 'update alias')

        # Ensure a user can't inadvertently update a TargetName for a different Target
        invalid_data = {
            'aliases': [
                {'id': alias.id, 'name': 'alias for wrong target'}
            ]
        }
        response = self.client.patch(reverse('api:targets-detail', args=(self.nst.id,)), data=invalid_data)
        self.assertContains(response,
                            f'TargetName identified by id \'{alias.id}\' is not an',
                            status_code=status.HTTP_400_BAD_REQUEST)

        # Ensure proper exception handling when updating or creating with the same name
        updates = {
            'aliases': [
                {'name': self.st.name}
            ]
        }
        response = self.client.patch(reverse('api:targets-detail', args=(self.st.id,)), data=updates)
        self.assertContains(response,
                            f'Alias \'{self.st.name}\' conflicts with Target name \'{self.st.name}\'',
                            status_code=status.HTTP_400_BAD_REQUEST)

    def test_target_delete(self):
        response = self.client.delete(reverse('api:targets-detail', args=(self.st.id,)))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(Target.objects.filter(pk=self.st.id).exists())

        self.client.force_login(self.user2)
        response = self.client.delete(reverse('api:targets-detail', args=(self.nst.id,)))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class TestTargetNameViewset(APITestCase):
    def setUp(self):
        user = User.objects.create(username='testuser')
        self.st = SiderealTargetFactory.create()
        self.alias = TargetNameFactory.create(target=self.st)
        assign_perm('tom_targets.view_target', user, self.st)
        assign_perm('tom_targets.delete_target', user, self.st)

        self.user2 = User.objects.create(username='testuser2')

        self.client.force_login(user)

    def test_targetname_detail(self):
        response = self.client.get(reverse('api:targetname-detail', args=(self.alias.id,)))
        self.assertEqual(response.json()['name'], self.alias.name)

        # Ensure that a user without view_target permission cannot access the target name
        self.client.force_login(self.user2)
        response = self.client.get(reverse('api:targetname-detail', args=(self.alias.id,)))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_targetname_delete(self):
        response = self.client.delete(reverse('api:targetname-detail', args=(self.alias.id,)))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(TargetName.objects.filter(pk=self.alias.id).exists())


class TestTargetExtraViewset(APITestCase):
    def setUp(self):
        user = User.objects.create(username='testuser')
        self.st = SiderealTargetFactory.create()
        self.extra = TargetExtraFactory.create(target=self.st)
        assign_perm('tom_targets.view_target', user, self.st)
        assign_perm('tom_targets.delete_target', user, self.st)

        self.user2 = User.objects.create(username='testuser2')

        self.client.force_login(user)

    def test_targetextra_detail(self):
        response = self.client.get(reverse('api:targetextra-detail', args=(self.extra.id,)))
        self.assertEqual(response.json()['id'], self.extra.id)

        # Ensure that a user without view_target permission cannot access the target extra
        self.client.force_login(self.user2)
        response = self.client.get(reverse('api:targetextra-detail', args=(self.extra.id,)))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_targetextra_delete(self):
        response = self.client.delete(reverse('api:targetextra-detail', args=(self.extra.id,)))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(TargetExtra.objects.filter(pk=self.extra.id).exists())



from datetime import datetime, timedelta
from urllib.parse import urlencode

from django import forms, template
from django.conf import settings
from django.urls import reverse
from guardian.shortcuts import get_objects_for_user
from plotly import offline
import plotly.graph_objs as go

from tom_observations.forms import AddExistingObservationForm, UpdateObservationId
from tom_observations.models import ObservationRecord
from tom_observations.facility import get_service_class, get_service_classes
from tom_observations.observation_template import ApplyObservationTemplateForm
from tom_observations.utils import get_sidereal_visibility
from tom_targets.models import Target


register = template.Library()


@register.inclusion_tag('tom_observations/partials/update_status_button.html', takes_context=True)
def update_status_button(context):
    """
    An inelegant way of passing filters in query parameters while updating observation statuses.
    Used in ObservationListView to retain filters.
    """
    return {'query_params': urlencode(context['request'].GET.dict())}


@register.filter
def display_obs_type(value):
    """
    This converts SAMPLE_TITLE into Sample Title. Used for display all-caps observation type in the
    tabs as titles.
    """
    return value.replace('_', ' ').title()


@register.inclusion_tag('tom_observations/partials/observing_buttons.html')
def observing_buttons(target):
    """
    Displays the observation buttons for all facilities available in the TOM.
    """
    facilities = get_service_classes()
    return {'target': target, 'facilities': facilities}


@register.inclusion_tag('tom_observations/partials/existing_observation_form.html')
def existing_observation_form(target):
    """
    Renders a form for adding an existing API-based observation to a Target.
    """
    return {'form': AddExistingObservationForm(initial={'target_id': target.id})}


@register.inclusion_tag('tom_observations/partials/update_observation_id_form.html')
def update_observation_id_form(obsr):
    """
    Renders a form for updating the observation ID for an ObservationRecord.
    """
    return {'form': UpdateObservationId(initial={'obsr_id': obsr.id, 'observation_id': obsr.observation_id})}


@register.inclusion_tag('tom_observations/partials/observation_type_tabs.html', takes_context=True)
def observation_type_tabs(context):
    """
    Displays tabs in observation creation form representing each available observation type.
    """
    request = context['request']
    query_params = request.GET.copy()
    observation_type = query_params.pop('observation_type', None)
    return {
        'params': urlencode(query_params),
        'type_choices': context['type_choices'],
        'observation_type': observation_type,
        'facility': context['form']['facility'].value,
        'target_id': request.GET.get('target_id')
    }


@register.inclusion_tag('tom_observations/partials/facility_observation_form.html')
def facility_observation_form(target, facility, observation_type):
    """
    Displays a form for submitting an observation for a specific facility and observation type, e.g., imaging.
    """
    facility_class = get_service_class(facility)()
    initial_fields = {
        'target_id': target.id,
        'facility': facility,
        'observation_type': observation_type
    }
    obs_form = facility_class.get_form(observation_type)(initial=initial_fields)
    obs_form.helper.form_action = reverse('tom_observations:create', kwargs={'facility': facility})

    return {'obs_form': obs_form}


@register.inclusion_tag('tom_observations/partials/observation_plan.html')
def observation_plan(target, facility, length=7, interval=60, airmass_limit=None):
    """
    Displays form and renders plot for visibility calculation. Using this templatetag to render a plot requires that
    the context of the parent view have values for start_time, end_time, and airmass.
    """

    visibility_graph = ''
    start_time = datetime.now()
    end_time = start_time + timedelta(days=length)

    visibility_data = get_sidereal_visibility(target, start_time, end_time, interval, airmass_limit)
    i = 0
    plot_data = []
    for site, data in visibility_data.items():
        plot_data.append(go.Scatter(x=data[0], y=data[1], mode='markers+lines', marker={'symbol': i}, name=site))
        i += 1
    layout = go.Layout(
        xaxis={'title': 'Date'},
        yaxis={'autorange': 'reversed', 'title': 'Airmass'}
    )
    visibility_graph = offline.plot(
        go.Figure(data=plot_data, layout=layout), output_type='div', show_link=False
    )

    return {
        'visibility_graph': visibility_graph
    }


@register.inclusion_tag('tom_observations/partials/observation_list.html', takes_context=True)
def observation_list(context, target=None):
    """
    Displays a list of all observations in the TOM, limited to an individual target if specified.
    """
    if target:
        if settings.TARGET_PERMISSIONS_ONLY:
            observations = target.observationrecord_set.all()
        else:
            observations = get_objects_for_user(
                                context['request'].user,
                                'tom_observations.view_observationrecord'
                            ).filter(target=target)
    else:
        observations = ObservationRecord.objects.all().order_by('-created')
    return {'observations': observations}


@register.inclusion_tag('tom_observations/partials/observationtemplate_run.html')
def observationtemplate_run(target):
    """
    Renders the form for running an observation template.
    """
    form = ApplyObservationTemplateForm(initial={'target': target})
    form.fields['target'].widget = forms.HiddenInput()
    return {'form': form}


@register.inclusion_tag('tom_observations/partials/observationtemplate_from_record.html')
def observationtemplate_from_record(obsr):
    """
    Renders a button that will pre-populate and observation template form with parameters from the specified
    ``ObservationRecord``.
    """
    obs_params = obsr.parameters
    obs_params.pop('target_id', None)
    template_params = urlencode(obs_params)
    return {
        'facility': obsr.facility,
        'params': template_params
    }


@register.inclusion_tag('tom_observations/partials/observation_distribution.html')
def observation_distribution(observations):
    """
    Displays a plot showing on a map the locations of all observations recorded in the TOM.
    """

    # "distinct" query is not supported, must manually find distinct observation per target
    sorted_observations = observations.order_by('scheduled_end')  # ascending so that only the max is preserved
    observation_targets = {}
    for obs in sorted_observations:
        observation_targets[obs.target_id] = (obs.status, obs.terminal)

    observation_no_status = [t for t in observation_targets.keys()
                             if not observation_targets[t][0]]  # status==""
    observation_terminal = [t for t in observation_targets.keys()
                            if observation_targets[t][0]
                            and observation_targets[t][1]]  # status!="" and terminal
    observation_non_terminal = [t for t in observation_targets.keys()
                                if observation_targets[t][0]
                                and not observation_targets[t][1]]  # status!="" and not terminal

    targets_no_status = Target.objects.filter(pk__in=observation_no_status)
    targets_terminal = Target.objects.filter(pk__in=observation_terminal)
    targets_non_terminal = Target.objects.filter(pk__in=observation_non_terminal)

    locations_no_status = targets_no_status.filter(type=Target.SIDEREAL).values_list('ra', 'dec', 'name')
    locations_terminal = targets_terminal.filter(type=Target.SIDEREAL).values_list('ra', 'dec', 'name')
    locations_non_terminal = targets_non_terminal.filter(type=Target.SIDEREAL).values_list('ra', 'dec', 'name')

    data = [
        dict(
            lon=[location[0] for location in locations_no_status],
            lat=[location[1] for location in locations_no_status],
            text=[location[2] for location in locations_no_status],
            hoverinfo='lon+lat+text',
            mode='markers',
            marker=dict(color='rgba(90, 90, 90, .8)'),
            type='scattergeo'
        ),
        dict(
            lon=[location[0] for location in locations_non_terminal],
            lat=[location[1] for location in locations_non_terminal],
            text=[location[2] for location in locations_non_terminal],
            hoverinfo='lon+lat+text',
            mode='markers',
            marker=dict(color='rgba(152, 0, 0, .8)'),
            type='scattergeo'
        ),
        dict(
            lon=[location[0] for location in locations_terminal],
            lat=[location[1] for location in locations_terminal],
            text=[location[2] for location in locations_terminal],
            hoverinfo='lon+lat+text',
            mode='markers',
            marker=dict(color='rgba(0, 152, 0, .8)'),
            type='scattergeo'
        ),
        dict(
            lon=list(range(0, 360, 60))+[180]*4,
            lat=[0]*6+[-60, -30, 30, 60],
            text=list(range(0, 360, 60))+[-60, -30, 30, 60],
            hoverinfo='none',
            mode='text',
            type='scattergeo'
        )
    ]
    layout = {
        'title': 'Observation Distribution (sidereal)',
        'hovermode': 'closest',
        'showlegend': False,
        'geo': {
            'projection': {
                'type': 'mollweide',
            },
            'showcoastlines': False,
            'showland': False,
            'lonaxis': {
                'showgrid': True,
                'range': [0, 360],
            },
            'lataxis': {
                'showgrid': True,
                'range': [-90, 90],
            },
        }
    }
    figure = offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)
    return {'figure': figure}


@register.inclusion_tag('tom_observations/partials/facility_status.html')
def facility_status():
    """
    Collect the facility status from the registered facilities and pass them
    to the facility_status.html partial template.
    See lco.py Facility implementation for example.
    :return:
    """

    facility_statuses = []
    for facility_class in get_service_classes().values():
        facility = facility_class()
        weather_urls = facility.get_facility_weather_urls()
        status = facility.get_facility_status()

        # add the weather_url to the site dictionary
        for site in status.get('sites', []):
            url = next((site_url['weather_url'] for site_url in weather_urls.get('sites', [])
                        if site_url['code'] == site['code']), None)
            if url is not None:
                site['weather_url'] = url

        facility_statuses.append(status)

    return {'facilities': facility_statuses}


@register.inclusion_tag('tom_observations/partials/facility_map.html')
def facility_map():
    facility_locations = []
    for facility_class in get_service_classes().values():
        facility = facility_class()
        sites = facility.get_observing_sites()

        # Flatten each facility site dictionary and add text label for use in facility map
        # Resulting list is of the format [['LCO', 'Siding Spring', 'coj', -31.272, 149.07, 1116], ...]
        facility_locations.extend([
            [facility.name, site_name] + [value for value in site_data.values()]
            for site_name, site_data in sites.items()
        ])

    data = [
        dict(
            lat=[site[3] for site in facility_locations],
            lon=[site[4] for site in facility_locations],
            text=[f'{site[0]}: {site[1]}' for site in facility_locations],
            hoverinfo='text',
            mode='markers',
            type='scattergeo'
        )
    ]
    layout = {
        'title': 'Facility Sites',
        'hovermode': 'closest',
        'showlegend': False,
        'geo': {
            'projection': {
                'type': 'mollweide',
            },
            'showcoastlines': False,
            'showland': True,
            'lonaxis': {
                'showgrid': True,
                'range': [0, 360],
            },
            'lataxis': {
                'showgrid': True,
                'range': [-90, 90],
            },
        }
    }
    figure = offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)
    return {'figure': figure}



# Generated by Django 2.1.4 on 2019-01-23 20:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0003_merge_20190123_1754'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='target',
            name='source',
        ),
        migrations.RemoveField(
            model_name='target',
            name='source_location',
        ),
        migrations.AlterField(
            model_name='target',
            name='name2',
            field=models.CharField(blank=True, default='', help_text='An alternative name for this target', max_length=100, verbose_name='Name 2'),
        ),
        migrations.AlterField(
            model_name='target',
            name='name3',
            field=models.CharField(blank=True, default='', help_text='An alternative name for this target', max_length=100, verbose_name='Name 3'),
        ),
    ]



from django.test import TestCase, override_settings
from django.contrib.auth.models import User
from django.urls import reverse

from tom_catalogs.harvester import AbstractHarvester, get_service_classes, MissingDataException


class TestHarvester(AbstractHarvester):
    name = 'TEST'

    def query(self, term):
        if term == 'notfound':
            raise MissingDataException
        self.catalog_data = {'ra': 24, 'dec': 77, 'name': 'faketarget', 'type': 'SIDEREAL'}

    def to_target(self):
        target = super().to_target()
        target.name = self.catalog_data['name']
        target.type = self.catalog_data['type']
        target.ra = self.catalog_data['ra']
        target.dec = self.catalog_data['dec']
        return target


@override_settings(TOM_HARVESTER_CLASSES=['tom_catalogs.tests.tests.TestHarvester'])
class TestHarvesterClass(TestCase):
    def test_get_broker_class(self):
        self.assertIn(TestHarvester, get_service_classes().values())


@override_settings(TOM_HARVESTER_CLASSES=['tom_catalogs.tests.tests.TestHarvester'])
class TestHarvesterViews(TestCase):
    def setUp(self):
        user = User.objects.create_user(username='test', password='test')
        self.client.force_login(user)

    def test_service_available(self):
        response = self.client.get(reverse('tom_catalogs:query'))
        self.assertContains(response, TestHarvester.name)

    def test_do_search(self):
        data = {'term': 'atarget', 'service': 'TEST'}
        response = self.client.post(reverse('tom_catalogs:query'), data=data, follow=True)
        self.assertContains(response, 'faketarget')

    def test_not_found(self):
        data = {'term': 'notfound', 'service': 'TEST'}
        response = self.client.post(reverse('tom_catalogs:query'), data=data, follow=True)
        self.assertContains(response, 'Object not found')



# Generated by Django 2.1.4 on 2019-05-03 23:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_observations', '0002_auto_20190306_2343'),
    ]

    operations = [
        migrations.AlterField(
            model_name='observationrecord',
            name='observation_id',
            field=models.CharField(max_length=255),
        ),
    ]



from datetime import datetime
from itertools import islice
import json
from requests import Response

from django.utils import timezone
from django.test import override_settings, tag, TestCase
from unittest import mock

from tom_alerts.brokers.mars import MARSBroker
from tom_alerts.alerts import get_service_class
from tom_targets.models import Target
from tom_dataproducts.models import ReducedDatum

alert1 = {
    'candid': 617122521615015023,
    'candidate': {
       'b': 0.70548695469711,
       'dec': -10.5296018,
       'jd': 2458371.6225231,
       'l': 20.7124513780029,
       'magpsf': 16.321626663208,
       'ra': 276.5843017,
       'rb': 0.990000009536743,
       'wall_time': 'Mon, 10 Sep 2018 02:56:25 GMT',
    },
    'lco_id': 11296149,
    'objectId': 'ZTF18abbkloa',
}


@override_settings(TOM_ALERT_CLASSES=['tom_alerts.brokers.mars.MARSBroker'])
class TestMARSBrokerClass(TestCase):
    """ Test the functionality of the MARSBroker, we modify the django settings to make sure
    it is the only installed broker.
    """
    def setUp(self):
        self.test_target = Target.objects.create(name='ZTF18aberpsh')
        ReducedDatum.objects.create(
            source_name='MARS',
            source_location=11053318,
            target=self.test_target,
            data_type='photometry',
            timestamp=timezone.now(),
            value=12
        )
        alert2 = alert1.copy()
        alert2['lco_id'] = 11053318
        alert2['objectId'] = 'ZTF18aberpsh'
        self.test_data = [alert1, alert2]

    def test_get_broker_class(self):
        self.assertEqual(MARSBroker, get_service_class('MARS'))

    def test_get_invalid_broker(self):
        with self.assertRaises(ImportError):
            get_service_class('LASAIR')

    @mock.patch('tom_alerts.brokers.mars.requests.get')
    def test_fetch_alerts(self, mock_requests_get):
        mock_return_data = {
            "has_next": "false",
            "has_prev": "false",
            "pages": 1,
            "results": [self.test_data[1]]
        }
        mock_response = Response()
        mock_response._content = str.encode(json.dumps(mock_return_data))
        mock_response.status_code = 200
        mock_requests_get.return_value = mock_response

        alerts = MARSBroker().fetch_alerts({'objectId': 'ZTF18aberpsh'})
        self.assertEqual(self.test_data[1]['objectId'], list(alerts)[0]['objectId'])

    def test_process_reduced_data_with_alert(self):
        test_alert = self.test_data[1]
        test_alert['prv_candidate'] = [
            {
                'candidate': {
                    'jd': 2458372.6225231,
                    'magpsf': 13,
                    'sigmapsf': 0.5,
                    'fid': 1
                }
            }
        ]

        MARSBroker().process_reduced_data(self.test_target, alert=test_alert)
        reduced_data = ReducedDatum.objects.filter(target=self.test_target, source_name='MARS')
        self.assertEqual(reduced_data.count(), 2)

    @mock.patch('tom_alerts.brokers.mars.MARSBroker.fetch_alert')
    def test_process_reduced_data_no_alert(self, mock_fetch_alert):
        self.test_data = self.test_data[1]
        self.test_data['prv_candidate'] = [
            {
                'candidate': {
                    'jd': 2458372.6225231,
                    'magpsf': 13,
                    'sigmapsf': 0.5,
                    'fid': 1
                }
            }
        ]
        mock_fetch_alert.return_value = self.test_data

        MARSBroker().process_reduced_data(self.test_target)
        reduced_data = ReducedDatum.objects.filter(target=self.test_target, source_name='MARS')
        self.assertEqual(reduced_data.count(), 2)

    def test_to_target(self):
        test_alert = self.test_data[0]

        created_target = MARSBroker().to_target(test_alert)
        self.assertEqual(created_target.name, 'ZTF18abbkloa')

    def test_to_generic_alert(self):
        test_alert = self.test_data[1]

        created_alert = MARSBroker().to_generic_alert(test_alert)
        self.assertEqual(created_alert.name, 'ZTF18aberpsh')


@tag('canary')
class TestMARSModuleCanary(TestCase):
    def setUp(self):
        self.broker = MARSBroker()
        self.expected_keys = ['avro', 'candid', 'candidate', 'lco_id', 'objectId', 'publisher']
        self.expected_candidate_keys = ['aimage', 'aimagerat', 'b', 'bimage', 'bimagerat', 'candid', 'chinr', 'chipsf',
                                        'classtar', 'clrcoeff', 'clrcounc', 'clrmed', 'clrrms', 'dec', 'decnr',
                                        'deltamaglatest', 'deltamagref', 'diffmaglim', 'distnr', 'distpsnr1',
                                        'distpsnr2', 'distpsnr3', 'drb', 'drbversion', 'dsdiff', 'dsnrms', 'elong',
                                        'exptime', 'fid', 'field', 'filter', 'fwhm', 'isdiffpos', 'jd', 'jdendhist',
                                        'jdendref', 'jdstarthist', 'jdstartref', 'l', 'magap', 'magapbig', 'magdiff',
                                        'magfromlim', 'maggaia', 'maggaiabright', 'magnr', 'magpsf', 'magzpsci',
                                        'magzpscirms', 'magzpsciunc', 'mindtoedge', 'nbad', 'ncovhist', 'ndethist',
                                        'neargaia', 'neargaiabright', 'nframesref', 'nid', 'nmatches', 'nmtchps',
                                        'nneg', 'objectidps1', 'objectidps2', 'objectidps3', 'pdiffimfilename', 'pid',
                                        'programid', 'programpi', 'ra', 'ranr', 'rb', 'rbversion', 'rcid', 'rfid',
                                        'scorr', 'seeratio', 'sgmag1', 'sgmag2', 'sgmag3', 'sgscore1', 'sgscore2',
                                        'sgscore3', 'sharpnr', 'sigmagap', 'sigmagapbig', 'sigmagnr', 'sigmapsf',
                                        'simag1', 'simag2', 'simag3', 'sky', 'srmag1', 'srmag2', 'srmag3', 'ssdistnr',
                                        'ssmagnr', 'ssnamenr', 'ssnrms', 'sumrat', 'szmag1', 'szmag2', 'szmag3',
                                        'tblid', 'tooflag', 'wall_time', 'xpos', 'ypos', 'zpclrcov', 'zpmed']

    def test_fetch_alerts(self):
        response = self.broker.fetch_alerts({'time__gt': '2018-06-01', 'time__lt': '2018-06-30'})

        alerts = []
        for alert in islice(response, 10):
            alerts.append(alert)
        self.assertEqual(len(alerts), 10)

        for key in self.expected_keys:
            self.assertTrue(key in alerts[0].keys())
        for key in self.expected_candidate_keys:
            self.assertTrue(key in alerts[0]['candidate'].keys())

    def test_fetch_alert(self):
        alert = self.broker.fetch_alert(1065519)

        for key in self.expected_keys:
            self.assertTrue(key in alert.keys())
        for key in self.expected_candidate_keys:
            self.assertTrue(key in alert['candidate'].keys())

    def test_process_reduced_data(self):
        alert = self.broker.fetch_alert(1065519)
        t = Target.objects.create(name='test target', ra=1, dec=2)
        self.broker.process_reduced_data(t, alert=alert)
        self.assertGreaterEqual(ReducedDatum.objects.filter(target=t, timestamp__lte=datetime(2020, 11, 3)).count(),
                                526)



# Generated by Django 2.1.4 on 2018-12-06 21:57

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='BrokerQuery',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('broker', models.CharField(max_length=50)),
                ('parameters', models.TextField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



# Generated by Django 2.1.4 on 2019-01-17 22:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='target',
            name='source',
            field=models.TextField(help_text='The original source reference for the target data.', max_length=100, null=True, verbose_name='Target Source'),
        ),
        migrations.AddField(
            model_name='target',
            name='source_location',
            field=models.TextField(help_text='URL or path to original target source reference.', max_length=100, null=True, verbose_name='Target Source Location'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-03-06 23:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_observations', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='observationrecord',
            name='scheduled_end',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='observationrecord',
            name='scheduled_start',
            field=models.DateTimeField(null=True),
        ),
    ]



from crispy_forms.layout import Layout, HTML

from tom_alerts.alerts import GenericBroker, GenericQueryForm, GenericAlert


class SCIMMAQueryForm(GenericQueryForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper.inputs.pop()
        self.helper.layout = Layout(
            HTML('''
                <p>
                This plugin is a stub for the SCIMMA plugin. In order to install the full plugin, please see the
                instructions <a href="https://github.com/TOMToolkit/tom_scimma">here</a>.
                </p>
            '''),
            HTML('''<a class="btn btn-outline-primary" href={% url 'tom_alerts:list' %}>Back</a>''')
        )


class SCIMMABroker(GenericBroker):
    name = 'SCIMMA'
    form = SCIMMAQueryForm

    def fetch_alerts(self, parameters):
        return iter([])

    def process_reduced_data(self, target, alert=None):
        return

    def to_generic_alert(self, alert):
        return GenericAlert()



from dateutil.parser import parse
import json
import re
import requests
from requests.exceptions import HTTPError

from astropy.coordinates import SkyCoord
from astropy.time import Time, TimezoneInfo
import astropy.units as u
from bs4 import BeautifulSoup
from django import forms

from tom_alerts.alerts import GenericAlert, GenericBroker, GenericQueryForm
from tom_dataproducts.models import ReducedDatum

BASE_BROKER_URL = 'http://gsaweb.ast.cam.ac.uk'


class GaiaQueryForm(GenericQueryForm):
    target_name = forms.CharField(required=False)
    cone = forms.CharField(
        required=False,
        label='Cone Search',
        help_text='RA,Dec,radius in degrees'
    )

    def clean_cone(self):
        cone = self.cleaned_data['cone']
        if cone:
            cone_params = cone.split(',')
            if len(cone_params) != 3:
                raise forms.ValidationError('Cone search parameters must be in the format \'RA,Dec,Radius\'.')
        return cone

    def clean(self):
        super().clean()
        if not (self.cleaned_data.get('target_name') or self.cleaned_data.get('cone')):
            raise forms.ValidationError('Please enter either a target name or cone search parameters.')
        elif self.cleaned_data.get('target_name') and self.cleaned_data.get('cone'):
            raise forms.ValidationError('Please only enter one of target name or cone search parameters.')


class GaiaBroker(GenericBroker):
    name = 'Gaia'
    form = GaiaQueryForm

    def fetch_alerts(self, parameters):
        """Must return an iterator"""
        response = requests.get(f'{BASE_BROKER_URL}/alerts/alertsindex')
        response.raise_for_status()

        soup = BeautifulSoup(response.content, 'html.parser')
        script_tags = soup.find_all('script')
        alerts = None

        alerts_pattern = re.compile(r'var alerts = \[(.*?)];')
        for script in script_tags:
            m = alerts_pattern.match(str(script.string).strip())
            if m is not None:
                alerts = '['+m.group(1)+']'
                break

        alert_list = json.loads(alerts)

        if parameters['cone'] is not None and len(parameters['cone']) > 0:
            cone_params = parameters['cone'].split(',')
            parameters['cone_ra'] = float(cone_params[0])
            parameters['cone_dec'] = float(cone_params[1])
            parameters['cone_radius'] = float(cone_params[2])*u.deg
            parameters['cone_centre'] = SkyCoord(float(cone_params[0]),
                                                 float(cone_params[1]),
                                                 frame="icrs", unit="deg")

        filtered_alerts = []
        if parameters.get('target_name'):
            for alert in alert_list:
                if parameters['target_name'] in alert['name']:
                    filtered_alerts.append(alert)

        elif 'cone_radius' in parameters.keys():
            for alert in alert_list:
                c = SkyCoord(float(alert['ra']), float(alert['dec']),
                             frame="icrs", unit="deg")
                if parameters['cone_centre'].separation(c) <= parameters['cone_radius']:
                    filtered_alerts.append(alert)

        else:
            filtered_alerts = alert_list

        return iter(filtered_alerts)

    def fetch_alert(self, target_name):

        alert_list = list(self.fetch_alerts({'target_name': target_name, 'cone': None}))

        if len(alert_list) == 1:
            return alert_list[0]
        else:
            return {}

    def to_generic_alert(self, alert):
        timestamp = parse(alert['obstime'])
        alert_link = alert.get('per_alert', {})['link']
        url = f'{BASE_BROKER_URL}/{alert_link}'

        return GenericAlert(
            timestamp=timestamp,
            url=url,
            id=alert['name'],
            name=alert['name'],
            ra=alert['ra'],
            dec=alert['dec'],
            mag=alert['alertMag'],
            score=1.0
        )

    def process_reduced_data(self, target, alert=None):

        if not alert:
            try:
                alert = self.fetch_alert(target.name)

            except HTTPError:
                raise Exception('Unable to retrieve alert information from broker')

        if alert is not None:
            alert_name = alert['name']
            alert_link = alert.get('per_alert', {})['link']
            lc_url = f'{BASE_BROKER_URL}/alerts/alert/{alert_name}/lightcurve.csv'
            alert_url = f'{BASE_BROKER_URL}/{alert_link}'
        elif target:
            lc_url = f'{BASE_BROKER_URL}/{target.name}/lightcurve.csv'
            alert_url = f'{BASE_BROKER_URL}/alerts/alert/{target.name}/'
        else:
            return

        response = requests.get(lc_url)
        response.raise_for_status()
        html_data = response.text.split('\n')

        for entry in html_data[2:]:
            phot_data = entry.split(',')

            if len(phot_data) == 3:
                if 'untrusted' not in phot_data[2] and 'null' not in phot_data[2]:
                    jd = Time(float(phot_data[1]), format='jd', scale='utc')
                    jd.to_datetime(timezone=TimezoneInfo())

                    value = {
                        'magnitude': float(phot_data[2]),
                        'filter': 'G'
                    }

                    rd, _ = ReducedDatum.objects.get_or_create(
                        timestamp=jd.to_datetime(timezone=TimezoneInfo()),
                        value=value,
                        source_name=self.name,
                        source_location=alert_url,
                        data_type='photometry',
                        target=target)
                    rd.save()

        return



# Generated by Django 2.1.4 on 2019-01-15 21:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='target',
            name='name2',
            field=models.CharField(default='', help_text='An alternative name for this target', max_length=100, verbose_name='Name'),
        ),
        migrations.AddField(
            model_name='target',
            name='name3',
            field=models.CharField(default='', help_text='An alternative name for this target', max_length=100, verbose_name='Name'),
        ),
    ]



from django.urls import path

from tom_observations.views import (AddExistingObservationView, FacilityStatusView, ObservationCreateView,
                                    ObservationRecordUpdateView, ObservationGroupCreateView, ObservationGroupDeleteView,
                                    ObservationGroupListView, ObservationListView, ObservationRecordCancelView,
                                    ObservationRecordDetailView, ObservationTemplateCreateView,
                                    ObservationTemplateDeleteView, ObservationTemplateListView,
                                    ObservationTemplateUpdateView)
from tom_observations.api_views import ObservationRecordViewSet
from tom_common.api_router import SharedAPIRootRouter

router = SharedAPIRootRouter()
router.register(r'observations', ObservationRecordViewSet, 'observations')

app_name = 'tom_observations'

urlpatterns = [
    path('add/', AddExistingObservationView.as_view(), name='add-existing'),
    path('list/', ObservationListView.as_view(), name='list'),
    path('status/', FacilityStatusView.as_view(), name='facility-status'),
    path('template/list/', ObservationTemplateListView.as_view(), name='template-list'),
    path('template/<str:facility>/create/', ObservationTemplateCreateView.as_view(), name='template-create'),
    path('template/<int:pk>/update/', ObservationTemplateUpdateView.as_view(), name='template-update'),
    path('template/<int:pk>/delete/', ObservationTemplateDeleteView.as_view(), name='template-delete'),
    path('template/<int:pk>/', ObservationTemplateUpdateView.as_view(), name='template-detail'),
    # This path must be above <str:facility>/create
    path('groups/create/', ObservationGroupCreateView.as_view(), name='group-create'),
    path('groups/list/', ObservationGroupListView.as_view(), name='group-list'),
    path('groups/<int:pk>/delete/', ObservationGroupDeleteView.as_view(), name='group-delete'),
    path('<str:facility>/create/', ObservationCreateView.as_view(), name='create'),
    path('<int:pk>/cancel/', ObservationRecordCancelView.as_view(), name='cancel'),
    path('<int:pk>/update/', ObservationRecordUpdateView.as_view(), name='update'),
    path('<int:pk>/', ObservationRecordDetailView.as_view(), name='detail'),
]
